/*
 * Video Capture Driver ( Video for Linux 1/2 )
 * for the Matrox Marvel G200 and Rainbow Runner-G series
 *
 * This module is an interface to the G100 and G200 video extension
 * registers.  
 *
 * Copyright (C) 1999  Ryan Drake <stiletto@mediaone.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

/* Changes: 000219 - gz: .) Changed the picture parameter range 
                            (like brightness) to signed char.
                         .) Implemented VIDIOCGAUDIO.
			 .) Changed input channel order (to fix xawtv).
			    (_incompatible_ with the current buz_tools !!!!)
*/

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/kernel.h>
#include <linux/major.h>
#include <linux/malloc.h>
#include <linux/mm.h>
#include <linux/pci.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/pgtable.h>
#include <asm/page.h>
#include <linux/sched.h>
#include <asm/segment.h>
#include <linux/types.h>
#include <linux/wrapper.h>

#if LINUX_VERSION_CODE >= 0x020300
#include <linux/spinlock.h>
#else
#include <asm/spinlock.h>
#endif

#include <linux/videodev.h>

#include <linux/version.h>
#include <asm/uaccess.h>

#if LINUX_VERSION_CODE >= 0x020300
#include <linux/i2c-old.h>
#else
#include <linux/i2c.h>
#endif
#include "zr36060.h"
#include "zr36057.h"
#include "mgavideo.h"
#include "ks0127.h"
#include "tuner.h"
#include "audiochip.h"
#include "videodev_mjpeg.h"
#include <linux/video_decoder.h>
#include <linux/video_encoder.h>

#define MIN(a,b) (((a)>(b))?(b):(a))
#define MAX(a,b) (((a)>(b))?(a):(b))

#if VIDEO_MAX_FRAME <= 32
#define   V4L_MAX_FRAME   32
#elif VIDEO_MAX_FRAME <= 64
#define   V4L_MAX_FRAME   64
#else
#error   "Too many video frame buffers to handle"
#endif
#define   V4L_MASK_FRAME   (V4L_MAX_FRAME - 1)

#define V4L_NAME      "V4L-mgavideo"   /* name of the driver */

#define V4L_ERR       KERN_ERR     V4L_NAME
#define V4L_DEBUG     KERN_INFO    V4L_NAME
#define V4L_INFO      KERN_INFO    V4L_NAME
#define V4L_WARNING   KERN_WARNING V4L_NAME

#define dprintk     if (debug) printk

static int debug = 0; /* insmod parameter */

#if LINUX_VERSION_CODE >= 0x020100
MODULE_PARM(debug,"i");
#endif


#define MGA_MAX 4
#define MGA_NAME "mga_v4l1cap"
#define MGA_MAX_WIDTH 65535
#define MGA_MIN_WIDTH 8
#define MGA_MAX_HEIGHT 65535
#define MGA_MIN_HEIGHT 8 

#define IOCTL_DEBUG(x) x

static int v4l_nbufs = 2;
static int v4l_bufsize = 128;	/* Everybody should be able to work with this setting */

struct mga_v4l1 {
	struct video_device	video_dev;
	struct mga_dev		*mga;
	int			id;
	int			initialized;
	int			user;
        int			mgacaps;
	int			input;		/* Input channel: 0 = Composite, 1 = S-VHS */
	int			norm;		/* Norm: VIDEO_MODE_PAL or VIDEO_MODEp_NTSC */
	unsigned long		freq;		/* Freq: Tuner frequency */

	/* Video for Linux parameters */
	struct video_picture picture;	/* Current picture params */
	struct video_buffer buffer;	/* Current buffer params */
	struct video_window window;	/* Current window params */
	int buffer_set, window_set;	/* Flags if the above structures are set */
	int video_interlace;	/* Image on screen is interlaced */

#if LINUX_VERSION_CODE >= 0x020300
   wait_queue_head_t   v4l_capq;      /* wait here for grab to finish */
#else
   struct wait_queue * v4l_capq;      /* wait here for grab to finish */
#endif

	int v4l_grab_frame;	/* Frame number being currently grabbed */
	int v4l_grab_seq;	/* Number of frames grabbed */
	int gwidth;		/* Width of current memory capture */
	int gheight;		/* Height of current memory capture */
	int gformat;		/* Format of ... */
	int gbpl;		/* byte per line of ... */

	/* V4L grab queue of frames pending */
	unsigned v4l_pend_head;
	unsigned v4l_pend_tail;
	int v4l_pend[V4L_MAX_FRAME];        

        int map_mjpeg_buffers;
};

static int mga_num = 0;
static struct mga_v4l1 mga_devs[MGA_MAX];

static int mga_open(struct video_device *dev, int flags)
{
	struct mga_v4l1* mv = (struct mga_v4l1*)dev;

	if(flags == 0 && mv->user == 0) {
		dprintk("mgacap: opening device\n");
		mv->user++;

#if LINUX_VERSION_CODE >= 0x020300
		init_waitqueue_head(&mv->v4l_capq);
#else
		mv->v4l_capq         = NULL;
#endif
		mv->mga->codec_head = 0;
		mv->mga->codec_tail = 0;

		MOD_INC_USE_COUNT;
		return 0;
	}

	return -EBUSY;
}

static void mga_close(struct video_device *dev)
{
	struct mga_v4l1* mv = (struct mga_v4l1*)dev;
	enum	buz_codec_mode mode;
	
	dprintk("mgacap: closing device\n");

	/* disable interrupts */
        mgavideo_ivsync_enable(mv->mga, 0);
	mgavideo_command_enable(mv->mga, 0);
	mgavideo_eoi_enable(mv->mga, 0);
	mgavideo_buffer_enable(mv->mga, 0);

	wake_up_interruptible(&mv->v4l_capq);
	
	mode = BUZ_MODE_IDLE;
	zr36060_ioctl(&mv->mga->zr36060, ZR36060_SET_MODE, &mode);
	zr36060_ioctl(&mv->mga->zr36060, ZR36060_CANBUFS, NULL);


	mv->user--;

	MOD_DEC_USE_COUNT;
}


static long mga_read(struct video_device *dev, char *buf, unsigned long count, int nonblock)
{
	return -EINVAL;
}

static long mga_write(struct video_device *dev, const char *buf, unsigned long count, int nonblock)
{
	return -EINVAL;
}

static int mga_ioctl(struct video_device *dev, unsigned int cmd, void *arg)
{
	struct mga_v4l1* mv = (struct mga_v4l1*)dev;

	switch (cmd) {

	case VIDIOCGCAP:
		{
			struct video_capability b;
			dprintk(KERN_DEBUG "mga_v4l1 ioctl VIDIOCGCAP\n");
			strncpy(b.name, mv->video_dev.name, sizeof(b.name));
			b.type = VID_TYPE_CAPTURE |
			    VID_TYPE_TUNER |
			    VID_TYPE_OVERLAY |
			    VID_TYPE_CHROMAKEY |
			  /* VID_TYPE_FRAMERAM | */
			    VID_TYPE_SCALES;
			/* theoretically we could also flag VID_TYPE_SUBCAPTURE
			   but this is not even implemented in the BTTV driver */
			b.channels = 3;
			b.audios = 1;
			b.maxwidth = MGA_MAX_WIDTH;
			b.maxheight = MGA_MAX_HEIGHT;
			b.minwidth = MGA_MIN_WIDTH;
			b.minheight = MGA_MIN_HEIGHT;

			if (copy_to_user(arg, &b, sizeof(b))) {
				return -EFAULT;
			}
			return 0;
		}

	case VIDIOCGCHAN:
		{
			struct video_channel v;

			if (copy_from_user(&v, arg, sizeof(v))) {
				return -EFAULT;
			}
			dprintk(KERN_DEBUG "mga_v4l1 ioctl VIDIOCGCHAN for channel %d\n", v.channel);
			switch (v.channel) {
			case 0:
				strcpy(v.name, "Composite");
				v.tuners = 0;
				v.flags = 0;
				v.type = VIDEO_TYPE_CAMERA;
				break;
			case 1:
				strcpy(v.name, "S-Video");
				v.tuners = 0;
				v.flags = 0;
				v.type = VIDEO_TYPE_CAMERA;
				break;

			case 2:
				strcpy(v.name, "Television");
				v.tuners = 1;
				v.flags = VIDEO_VC_TUNER;
				v.type = VIDEO_TYPE_TV;
				break;

#ifdef	NEVER
	/* No tuner cards for mga have this input connected */
			case 3:
				strcpy(v.name, "YUV656 60Hz");
				v.tuners = 0;
				v.flags = 0;
				v.type = VIDEO_TYPE_CAMERA;
				break;
			case 4:
				strcpy(v.name, "YUV656 50Hz");
				v.tuners = 0;
				v.flags = 0;
				v.type = VIDEO_TYPE_CAMERA;
				break;
#endif
			default:
				return -EINVAL;
			}
			v.norm = mv->norm;

			if(mv->mgacaps & MGAVIDEO_HAS_AUDIO) {
				v.flags |= VIDEO_VC_AUDIO;
			}

			if (copy_to_user(arg, &v, sizeof(v))) {
				return -EFAULT;
			}
			return 0;
		}

	case VIDIOCSCHAN:
		{
			struct video_channel v;
			int input, norm;

			if (copy_from_user(&v, arg, sizeof(v))) {
				return -EFAULT;
			}
			dprintk(KERN_DEBUG "mga_v4l1 ioctl VIDIOCSCHAN: channel=%d, norm=%d\n", 
			       v.channel, v.norm);
			switch (v.channel) {
			case 0:
				input = KS_INPUT_COMPOSITE;
				break;
			case 1:
				input = KS_INPUT_SVIDEO;
				break;

			case 2:
				input = KS_INPUT_TUNER;
				break;
#ifdef	NEVER
	/* No tuner cards for mga have this input connected */
			case 3:
				input = KS_INPUT_YUV656_60HZ;
				break;
			case 4:
				input = KS_INPUT_YUV656_50HZ;
				break;
#endif
			default:
				return -EINVAL;
			}

			switch (v.norm) {
			case VIDEO_MODE_PAL:
				norm = KS_STD_PAL;
				break;
			case VIDEO_MODE_NTSC:
				norm = KS_STD_NTSC;
				break;
			default:
				return -EINVAL;
			}

			mv->input = input;
			mv->norm = norm;

			mgavideo_decoder( mv->mga, KS0127_SET_INPUT, &input );
			mgavideo_decoder( mv->mga, KS0127_SET_STANDARD, &norm );

			return 0;
		}

	case VIDIOCGTUNER:
		{
			struct video_tuner v;
			if(copy_from_user(&v, arg, sizeof(v)) != 0)
				return -EFAULT;
			if(v.tuner != 0)
				return -EINVAL;
			strcpy(v.name, "Television");
			v.rangelow = 0;
			v.rangehigh = 0xFFFFFFFF;
			v.flags = VIDEO_TUNER_PAL |
			          VIDEO_TUNER_NTSC |
				  VIDEO_TUNER_SECAM;
			v.mode = mv->norm;
			v.signal = 0; /* unknown ? should show if the tuner has catched a signal */
			if(copy_to_user(arg, &v, sizeof(v)))
				return -EFAULT;
			return 0;
		}

	case VIDIOCSTUNER:
		{
			struct video_tuner v;
			dprintk("mga_v4l1 ioctl VIDIOCSTUNER v.tuner=%d\n", v.tuner);
			if(copy_from_user(&v, arg, sizeof(v)))
				return -EFAULT;
			/* Only one channel has a tuner */
			if(v.tuner != 0) /* The tuner is on channel 2, not 0 */
				return -EINVAL;
			if(v.mode!=VIDEO_MODE_PAL&&v.mode!=VIDEO_MODE_NTSC
			   &&v.mode!=VIDEO_MODE_SECAM)
				return -EOPNOTSUPP;

			if (mv->norm != v.mode) 
			  {
			    mgavideo_tuner(mv->mga, TUNER_SET_MODE,&v.mode);
			    mv->norm = v.mode; 
			  }
			return 0;
		}

	case VIDIOCGPICT:
		{
			struct video_picture p = mv->picture;

			dprintk(KERN_DEBUG "mga_v4l1 ioctl VIDIOCGPICT\n");
			p.depth = mv->buffer.depth;
			switch (mv->buffer.depth) {
			case 15:
				p.palette = VIDEO_PALETTE_RGB555;
				break;

			case 16:
				p.palette = VIDEO_PALETTE_RGB565;
				break;

			case 24:
				p.palette = VIDEO_PALETTE_RGB24;
				break;

			case 32:
				p.palette = VIDEO_PALETTE_RGB32;
				break;
			}

			if (copy_to_user(arg, &p, sizeof(p))) {
				return -EFAULT;
			}
			return 0;
		}

	case VIDIOCSPICT:
		{
		        int brightness, hue, contrast, colour;
			struct video_picture p;

			dprintk("mgacap ioctl VIDIOCSPICT bri=%d hue=%d col=%d con=%d dep=%d pal=%d\n",
			       p.brightness, p.hue, p.colour, p.contrast, p.depth, p.palette);
			if (copy_from_user(&p, arg, sizeof(p))) {
				return -EFAULT;
			}
			/* We want -127 to 128 (signed char), but we get 0-65535 */  
			brightness = (p.brightness >> 8) - 128;
			hue = (p.hue >> 8) - 128;
			contrast = (p.contrast >> 8) - 128;
			colour = (p.colour >> 8) - 128;
			mgavideo_decoder( mv->mga, KS0127_SET_BRIGHTNESS, &brightness);
			mgavideo_decoder( mv->mga, KS0127_SET_HUE, &hue);
			mgavideo_decoder( mv->mga, KS0127_SET_SATURATION, &colour);
			mgavideo_decoder( mv->mga, KS0127_SET_CONTRAST, &contrast);

			mv->picture = p;
			return 0;
		}

	case VIDIOCCAPTURE:
		{
			int v;

			if (copy_from_user(&v, arg, sizeof(v))) {
				return -EFAULT;
			}
			printk(KERN_DEBUG "mga_v4l1 ioctl VIDIOCCAPTURE: %d\n", v);
			mgavideo_ivsync_enable( mv->mga, 1);
			mgavideo_preview_enable(mv->mga, v);
			return 0;
		}

	case VIDIOCGWIN:
		{
			struct video_window vw = mv->window;

			printk(KERN_DEBUG "mga_v4l1 ioctl VIDIOCGWIN\n");
			if (copy_to_user(arg, &vw, sizeof(vw))) 
				return -EFAULT;
			
			return 0;
		}

	case VIDIOCSWIN:
		{
			struct video_window vw;

			if (copy_from_user(&vw, arg, sizeof(vw))) {
				return -EFAULT;
			}
			dprintk("mga_v4l1 ioctl VIDIOCSWIN: x=%d y=%d w=%d h=%d chromakey=%d\n", 
			       vw.x, vw.y, vw.width, vw.height, vw.chromakey);

			mv->window = vw;
			mv->window_set = 1;

			mgavideo_set_window(mv->mga, vw.x, vw.y, vw.width, vw.height);
			mgavideo_set_colorkey(mv->mga, (vw.chromakey & 0xff0000) >> 16,
							(vw.chromakey & 0xff00) >> 8,
							vw.chromakey & 0xff);
			return 0;
		}

	case VIDIOCGFBUF:
		{
			dprintk(KERN_DEBUG "mga_v4l1 ioctl VIDIOCGFBUF\n");
			if (copy_to_user(arg, &mv->buffer, sizeof(mv->buffer))) {
				return -EFAULT;
			}
			return 0;
		}

	case VIDIOCSFBUF:
		{
			struct video_buffer v;

/* drake: TODO why do we check these here? they seem to prevent me from
 *        using v4l-conf and xawtv!
 */
#if 0
         /* RJ: Isn't this too restrictive? As long as the user doesn't set
                the base address it shouldn't be too dangerous */
         if (!capable(CAP_SYS_ADMIN)) {
            printk(BUZ_ERR ": Only the superuser may issue VIDIOCSFBUF ioctl\n");
            return -EPERM;
         }
#endif

			if (copy_from_user(&v, arg, sizeof(v)))
				return -EFAULT;
			
			dprintk("mga_v4l1 ioctl VIDIOCSFBUF: base=0x%x w=%d h=%d depth=%d bpl=%d\n", 
			       (u32) v.base, v.width, v.height, v.depth, v.bytesperline);

			if (v.depth != 15
			    && v.depth != 16
			    && v.depth != 24
			    && v.depth != 32) 
				return -EINVAL;
			
			if (v.height <= 0 || v.width <= 0 || v.bytesperline <= 0) 
				return -EINVAL;
			
			if (v.bytesperline & 3) 
				return -EINVAL;
			
			if (v.base) 
				mv->buffer.base = (void *) ((unsigned long) v.base & ~3);
			
			mv->buffer.height = v.height;
			mv->buffer.width = v.width;
			mv->buffer.depth = v.depth;
			mv->buffer.bytesperline = v.bytesperline;

			if (mv->buffer.base)
				mv->buffer_set = 1;
			mv->window_set = 0;	/* The user should set new window parameters */
			return 0;
		}

		/* RJ: what is VIDIOCKEY intended to do ??? */
		/* gz:it is there to support colorkeying, I think */

	case VIDIOCGFREQ:
		{
			if(copy_to_user(arg, &mv->freq, sizeof(mv->freq)))
				return -EFAULT;
			dprintk("mga_v4l1 ioctl VIDIOCGFREQ: freq=0x%lx\n", mv->freq);
			return 0;
		}

	case VIDIOCSFREQ:
		{
			if(copy_from_user(&mv->freq, arg, sizeof(mv->freq)))
				return -EFAULT;
			dprintk("mga_v4l1 ioctl VIDIOCSFREQ: freq=0x%lx\n", mv->freq);

			mgavideo_audio( mv->mga, AUDC_SWITCH_MUTE, 0 );
			mgavideo_tuner( mv->mga, TUNER_SET_TVFREQ, &mv->freq );
			mgavideo_audio( mv->mga, AUDC_SET_TVNORM, &mv->norm );
			mgavideo_audio( mv->mga, AUDC_NEWCHANNEL, 0 );

			return 0;
		}

	case VIDIOCGAUDIO:
        	{
		  struct video_audio v; 
		  dprintk(KERN_DEBUG "mga_v4l1 ioctl VIDIOCGAUDIO:  now supported !\n");
		  /*return -EINVAL;*/ 
		  /* v = mv->mga->audio_dev; Ok, let's try without, but I think I have to create one */
		  v.flags&=~(VIDEO_AUDIO_MUTE|VIDEO_AUDIO_MUTABLE);
		  v.flags|=VIDEO_AUDIO_MUTABLE;
		  strcpy(v.name,"Television");
		  
		  if (mv->mga->hasaudio) {
		    __u16 temp_voll,temp_volr;
		    v.flags|=VIDEO_AUDIO_VOLUME |
		      VIDEO_AUDIO_BASS |
		      VIDEO_AUDIO_TREBLE;
		    mgavideo_audio( mv->mga, AUDC_GET_VOLUME_LEFT, &temp_voll);
		    mgavideo_audio( mv->mga, AUDC_GET_VOLUME_RIGHT, &temp_volr);
		    v.volume=MAX(temp_voll,temp_volr);
		    v.balance=(32768*MIN(temp_voll,temp_volr))/
		      (v.volume ? v.volume : 1);
		    
		    v.balance=(temp_voll<temp_volr)?
		      (65535-v.balance):v.balance;
		    mgavideo_audio( mv->mga,AUDC_GET_BASS,&(v.bass));
		    mgavideo_audio( mv->mga,AUDC_GET_TREBLE,&(v.treble));
		    mgavideo_audio( mv->mga,AUDC_GET_STEREO,&(v.mode));
		  }
		  else v.mode = VIDEO_SOUND_MONO; /* We would have no sound, how do I write that ? */
		  if(copy_to_user(arg,&v,sizeof(v)))
		    return -EFAULT;
		  return 0;
        	}
	case VIDIOCSAUDIO:
        	{
			struct video_audio v;
			dprintk("mga_v4l1 ioctl VIDIOCSAUDIO: vol=%d, bal=%d !\n",
				v.volume, v.balance);
			if(copy_from_user(&v,arg, sizeof(v)))
				return -EFAULT;

			if (mv->mga->hasaudio)
			  { if (!(v.flags&VIDEO_AUDIO_MUTE)) 
			    {
			      __u16 temp_vol;
			      temp_vol=(2*MIN(65536-v.balance,32768) *
					v.volume)/65536;
			      mgavideo_audio( mv->mga, AUDC_SET_VOLUME_LEFT, &temp_vol);	
			      temp_vol=(2*MIN(v.balance,32768) *
					v.volume)/65536;
			      mgavideo_audio( mv->mga, AUDC_SET_VOLUME_RIGHT, &temp_vol);
			      /* bass and treble are meaningless, but we set them anyways */
			      mgavideo_audio( mv->mga, AUDC_SET_BASS,&(v.bass));
			      mgavideo_audio( mv->mga, AUDC_SET_TREBLE,&(v.treble));
			      
			      mgavideo_audio( mv->mga, AUDC_SET_STEREO,&(v.mode));
			    }
			  else
			    mgavideo_audio( mv->mga, AUDC_SWITCH_MUTE, 0);	
			  }
			return 0;
        	}
		
	case VIDIOCSYNC:
		{
			int v;

			if (copy_from_user(&v, arg, sizeof(v))) 
				return -EFAULT;
			
			dprintk("mga_v4l1 ioctl VIDIOCSYNC %d\n", v);
			return 0;
		}

	case VIDIOCMCAPTURE:
		{
			struct video_mmap vm;

			if (copy_from_user((void *) &vm, (void *) arg, sizeof(vm))) {
				return -EFAULT;
			}
			dprintk(KERN_DEBUG "mga_v4l1 ioctl VIDIOCMCAPTURE frame=%d geom=%dx%d fmt=%d\n",
			       vm.frame, vm.height, vm.width, vm.format);
			return 0;
		}

	case VIDIOCGMBUF:
		{
			struct video_mbuf vm;
			int i;

			dprintk(KERN_DEBUG "mga_v4l1 ioctl VIDIOCGMBUF\n");
		
			vm.size = v4l_nbufs * v4l_bufsize;
			vm.frames = v4l_nbufs;
			for (i = 0; i < v4l_nbufs; i++) {
				vm.offsets[i] = i * v4l_bufsize;
			}

			if (copy_to_user(arg, &vm, sizeof(vm))) {
				return -EFAULT;
			}
			return 0;
		}

	case VIDIOCGUNIT:
		{
			struct video_unit vu;

			dprintk(KERN_DEBUG "mga_v4l1 ioctl VIDIOCGUNIT\n");
			vu.video = mv->video_dev.minor;
			vu.vbi = VIDEO_NO_UNIT;
			vu.radio = VIDEO_NO_UNIT;
			vu.audio = VIDEO_NO_UNIT;
			vu.teletext = VIDEO_NO_UNIT;
			if (copy_to_user(arg, &vu, sizeof(vu)))
				return -EFAULT;
			return 0;
		}

	case VIDIOCGCAPTURE:
        	{
			dprintk(KERN_DEBUG "mga_v4l1 ioctl VIDIOCGCAPTURE: unsupported\n");
			return -EINVAL;
        	}
	case VIDIOCSCAPTURE:
        	{
			dprintk(KERN_DEBUG "mga_v4l1 ioctl VIDIOCSCAPTURE: unsupported\n");
			return -EINVAL;
        	}

	case MJPIOC_G_PARAMS:
	  {
		struct buz_params	params;
		int			res;


		IOCTL_DEBUG(printk(V4L_DEBUG "%-u: ioctl MJPIOC_G_PARAMS\n", mv->id));

		res = zr36060_ioctl(&mv->mga->zr36060, ZR36060_GET_PARAMS, &params);
		if (res)
			return res;

		if (copy_to_user(arg, &params, sizeof(params))) 
			return -EFAULT;

		return 0;
	  }
	  
	case MJPIOC_S_PARAMS:
	  {
		int	res;
		struct buz_params params;

		if (copy_from_user(&params, arg, sizeof(params)))
			return -EFAULT;

		IOCTL_DEBUG(printk(V4L_DEBUG "-%u: ioctl MJPIOC_S_PARAMS\n", mv->id);)

		/* Check the params first before overwriting our internal values */
		res = zr36060_ioctl(&mv->mga->zr36060, ZR36060_SET_PARAMS, &params);
		if (res)
			return res;

		mv->input = params.input;
		mv->norm  = params.norm;
		

		/* Make changes of input and norm go into effect immediatly */
#ifdef	NEVER		
		i2c_control_device(&mv->mga->i2c, I2C_DRIVERID_VIDEODECODER, DECODER_SET_INPUT, &mv->input);
		i2c_control_device(&mv->mga->i2c, I2C_DRIVERID_VIDEODECODER, DECODER_SET_NORM, &mv->norm);
		i2c_control_device(&mv->mga->i2c, I2C_DRIVERID_VIDEOENCODER, ENCODER_SET_NORM, &mv->norm);
#endif
		if (copy_to_user(arg, &params, sizeof(params)))
			return -EFAULT;
		return 0;
	  }
	  
        case MJPIOC_REQBUFS:
	  {
		int	res;
		struct buz_requestbuffers br;

		if (copy_from_user(&br, arg, sizeof(br)))
			return -EFAULT;

		IOCTL_DEBUG(printk(V4L_DEBUG "-%u: ioctl MJPIOC_REQBUFS count = %lu size=%lu\n",
                            mv->id, br.count,br.size);)

		res = zr36060_ioctl(&mv->mga->zr36060, ZR36060_REQBUFS, &br);

		if (copy_to_user(arg, &br, sizeof(br)))
			return -EFAULT;

		return 0;
	  }

	case MJPIOC_QBUF_CAPT:
	  {
	    int nb;
	    
	    if (copy_from_user((void *)&nb, (void *)arg, sizeof(nb))) {
	      return -EFAULT;
	    }
	    IOCTL_DEBUG(printk(V4L_DEBUG "-%u: ioctl MJPIOC_QBUF_CAPT %d\n",mv->id, nb));

	    return zr36060_ioctl(&mv->mga->zr36060, ZR36060_QBUF_CAPT, &nb);
	  }
	  
	case MJPIOC_QBUF_PLAY:
	  {
	    int nb;
	    
	    if (copy_from_user((void *)&nb, (void *)arg, sizeof(nb))) 
	      return -EFAULT;
         
	    IOCTL_DEBUG(printk(V4L_DEBUG "-%u: ioctl BUZIOC_QBUF_PLAY %d\n",mv->id, nb));

	    return zr36060_ioctl(&mv->mga->zr36060, ZR36060_QBUF_PLAY, &nb);
	  }

	case MJPIOC_SYNC:
	  {
	    struct mjpeg_sync bs;
	    int res;
	    
	    IOCTL_DEBUG(printk(V4L_DEBUG "-%u: buz ioctl BUZIOC_SYNC\n", mv->id));

	    res =  zr36060_ioctl(&mv->mga->zr36060, ZR36060_SYNC, &bs);

	    if (copy_to_user(arg, &bs, sizeof(bs))) 
	      return -EFAULT;
	    
	    return res;
	  }
	  
	case MJPIOC_G_STATUS:
	  {
#ifdef	NEVER
	    struct mjpeg_status bs;
	    IOCTL_DEBUG(printk("mgavideo ioctl BUZIOC_G_STATUS\n"));

	    if (copy_from_user(&bs, arg, sizeof(bs))) 
	      return -EFAULT;

	    mgavideo_get_channelstatus(mv->mga, bs.input, &bs.signal, &bs.norm, &bs.color);
	    
	    if (copy_to_user(arg, &bs, sizeof(bs))) 
	      return -EFAULT;
#endif	    
	    return 0;
	  }
	default:
		return -ENOIOCTLCMD;

	}
	return 0;
}


static int mga_mmap(struct video_device *dev, const char *adr, unsigned long size)
{
	struct mga_v4l1 *mv = (struct mga_v4l1*)dev;
	unsigned long start = (unsigned long)adr;
	unsigned long page, todo;
	int i;
	int stamp[2];
	int field[2];
	__u8* input_buffer[2];

	if (mv->map_mjpeg_buffers)
	  {
	    if ((i = zr36060_mmap_bufs(&mv->mga->zr36060, (unsigned long)adr, size)))
	      return i;
	  }
	else
	  {
	    /* Map the V4L buffers */
	    if (size > v4l_nbufs * v4l_bufsize) 
	      return -EINVAL;
#ifdef	NEVER	    
	    mgavideo_lock_video(mv->mga, stamp, field, input_buffer);
#endif	    
	    for (i = 0; i < v4l_nbufs; i++) 
	      {
		todo = size;
		if (todo > v4l_bufsize)
		  todo = v4l_bufsize;
		
		page = (unsigned long)input_buffer[i];
		
		if (remap_page_range(start, page, todo, PAGE_SHARED)) {
		  printk(KERN_ERR "%s: mga_mmap(V4L): remap_page_range failed\n", mv->video_dev.name);
		  return -EAGAIN;
		}
		size -= todo;
		start += todo;
		if (size == 0)
			break;
	      }
#ifdef	NEVER
	    mgavideo_unlock_video(mv->mga, stamp, 0);
#endif
	    }
	
	    return 0;
}

static int mga_init_done(struct video_device *dev)
{
	return 0;
}

static struct video_device mga_template =
{
	MGA_NAME,
	VID_TYPE_CAPTURE | VID_TYPE_OVERLAY | VID_TYPE_CLIPPING |
	VID_TYPE_FRAMERAM | VID_TYPE_SCALES | VID_TYPE_SUBCAPTURE,
	VID_HARDWARE_BT848,
	mga_open,
	mga_close,
	mga_read,
	mga_write,
	NULL,
	mga_ioctl,
	mga_mmap,
	mga_init_done,
	NULL,
	0,
	0
};



static void mga_init(struct mga_v4l1* mv, struct mga_dev* mga)
{
	int	in, input, norm;
	struct buz_params codec_params;

	mv->mga = mga;
	mv->user = 0;
        mv->mgacaps = mgavideo_querycaps(mv->mga);

	mv->buffer_set = 0;
	mv->buffer.base = NULL;
	mv->buffer.width = 0;
	mv->buffer.height = 0;
	mv->buffer.depth = 0;
	mv->buffer.bytesperline = 0;

	mv->norm = 0;
	mv->input = 0;
        mv->freq = 0;
	mv->video_interlace = 0;
	mv->map_mjpeg_buffers = 1;

	mv->picture.colour = 32768;
	mv->picture.brightness = 32768;
	mv->picture.hue = 32768;
	mv->picture.contrast = 32768;
	mv->picture.whiteness = 0;
	mv->picture.depth = 0;
	mv->picture.palette = 0;

	/* set some sane default values */
        mgavideo_decoder( mga, KS0127_RESET, 0 );
	input = 0;
	codec_params.input = input;
        in = KS_INPUT_COMPOSITE;
        mgavideo_decoder( mga, KS0127_SET_INPUT, &in );
	norm = 0;
	codec_params.input = norm;
        in = KS_STD_NTSC;
        mgavideo_decoder( mga, KS0127_SET_STANDARD, &in );
        in = KS_OUTPUT_YUV656E;
        mgavideo_decoder( mga, KS0127_SET_OUTPUT, &in );
        in = -1;
        mgavideo_tuner( mga, TUNER_SET_TYPE, &in );

	mgavideo_zr36060_attach(mv->mga);

	mv->initialized = 1;
}

#ifdef MODULE
int init_module(void)
#else
int init_mga_cards(struct video_init *unused)
#endif
{
	struct mga_v4l1* mv;
	struct mga_dev* mga;

	printk(KERN_INFO "Matrox Marvel and Rainbow Runner G Capture Driver (V4L)\n");

        memset(mga_devs, 0, sizeof(*mga_devs) * MGA_MAX);

        while((mga = mgavideo_get()) && (mga_num < MGA_MAX)) {

		mv = &(mga_devs[mga_num]);

		memcpy(&mv->video_dev, &mga_template, sizeof(mga_template));
		sprintf(mv->video_dev.name, "mgav4l1%u", mga_num);

		
		if (video_register_device(&mv->video_dev, VFL_TYPE_GRABBER) < 0) {
			mgavideo_release(mga);
			return -EIO;
		}

		mga_init(mv, mga);
		mga_num++;
	}

	if(mga_num == 0) {
		return -ENXIO;
	}

	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{
	int i = 0;
	struct mga_v4l1* mv;

        while(i < MGA_MAX) {

		mv = &(mga_devs[i]);
		if(mv->initialized) {

			video_unregister_device(&mv->video_dev);

			mgavideo_release(mv->mga);

			mv->mga = NULL;
			mv->initialized = 0;
			mv->user = 0;

			mga_num--;
		}
                i++;
	}
}
#endif

#if 0 /* ring buffer for cutting in buz stuff */
#endif
