/*
 * Video Capture Driver ( Video for Linux 1/2 )
 * for the Matrox Marvel G200/G400 and Rainbow Runner-G series
 *
 * This module is an interface to the G100,G200 and G400 video extension
 * registers.  
 *
 * Copyright (C) 1999  Ryan Drake <stiletto@mediaone.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 *
 *****************************************************************************
 * Version History:
 * V1.0 Ryan Drake         Initial version by Ryan Drake
 * V1.1 Gerard v.d. Horst  Activated the backend scaler
 * V1.2 Niels de Carpentier Added G400 support
 */

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <linux/pci.h>
#include <linux/interrupt.h>
#if LINUX_VERSION_CODE >= 0x020300
#include <linux/i2c-old.h>
#else
#include <linux/i2c.h>
#endif
#include <linux/videodev.h>
#include <linux/wrapper.h>
#include "zr36060.h"
#include "mgavideo.h"
#include "tuner.h"
#include "audiochip.h"
#include "ks0127.h"
#include "tvnorm.h"
#ifndef	I2C_BUSID_MGA
#define I2C_BUSID_MGA 100
#endif


#define dprintk     if (debug) printk

static int debug = 0; /* insmod parameter */
static int memsize = 8192 * 1024;

#define MGA_NAME      "mgavideo"   /* name of the driver */

#define MGA_ERR       KERN_ERR     MGA_NAME
#define MGA_DEBUG     KERN_INFO    MGA_NAME
#define MGA_INFO      KERN_INFO    MGA_NAME
#define MGA_WARNING   KERN_WARNING MGA_NAME

#define DEBUG(x)           /* Debug driver */
#define IDEBUG(x)          /* Debug interrupt handler */
#define IOCTL_DEBUG(x)  

#if LINUX_VERSION_CODE >= 0x020100
MODULE_PARM(debug,"i");
MODULE_PARM(memsize,"i");
#endif

/* This is for <=2.2.x */
#ifndef PCI_DEVICE_ID_MATROX_G400
#define PCI_DEVICE_ID_MATROX_G400 0x0525
#endif

/* i2c registers */
#define PALWTADD        0x3c00
#define	PALRDADD	0x3c03
#define X_DATAREG       0x3c0a
#define	XMULCTRL	0x19
#define XMISCCTRL       0x1e
#define XGENIOCTRL      0x2a
#define XGENIODATA      0x2b
#define XKEYOPMODE      0x51
#define XCOLMSK0RED     0x52
#define XCOLMSK0GREEN   0x53
#define XCOLMSK0BLUE    0x54
#define XCOLKEY0RED     0x55
#define XCOLKEY0GREEN   0x56
#define XCOLKEY0BLUE    0x57
#define SDA             0x10
#define SCL             0x20

#define I2C_DELAY       10

/* Desktop depth values XMULCTRL register (p->palettized) */
#define MGA_DEPTH8p  0x00
#define MGA_DEPTH15  0x01
#define MGA_DEPTH16  0x02
#define MGA_DEPTH24  0x03
#define MGA_DEPTH32  0x04
#define MGA_DEPTH32p 0x07


static struct tvnorm tvnorms[] = {
   /* PAL-BDGHI */
   { 864, 720, 625, 576, 31, 16 , 31, 16},

   /* NTSC */
   { 858, 720, 525, 480, 60,  22, 60, 22 },
};


/****************************************************************************
* raw register access : these routines directly interact with the mga's
*                       control aperature.  must not be called until after
*                       the board's pci memory has been mapped.
****************************************************************************/
#ifdef	NEVER
static u8 _mgareadb(struct mga_dev* mga, u32 reg )
{
        return readb(mga->ctrl + reg );
}


static u16 _mgareadw(struct mga_dev* mga, u32 reg )
{
        return readw(mga->ctrl + reg );
}
#endif

static u32 _mgaread(struct mga_dev* mga, u32 reg )
{
        return readl(mga->ctrl + reg );
}

static void _mgawriteb( struct mga_dev* mga, u32 reg, u32 data )
{
        writeb( data, mga->ctrl + reg );
}

static void _mgawritew( struct mga_dev* mga, u32 reg, u32 data )
{
        writeb( data, mga->ctrl + reg );
}

static void _mgawrite( struct mga_dev* mga, u32 reg, u32 data )
{
        writel( data, mga->ctrl + reg );
}


/****************************************************************************
* mga i2c bus interface
****************************************************************************/
static void _mgai2c_attach( struct i2c_bus* bus, int id )
{
        struct mga_dev* mga = (struct mga_dev*)bus->data;
	int	i;

        dprintk( "mgavideo: i2c device attach: %d\n", id );

        switch( id )
        {
        case I2C_DRIVERID_MSP3400:
                mga->hasaudio = 1;
                break;

        case I2C_DRIVERID_TUNER:
                mga->hastuner = 1;
                break;

        case I2C_DRIVERID_VIDEODECODER:
                mga->hasdecoder = 1;
		for(i=0; i < bus->devcount; i++) {
			if (bus->devices[i]->addr == 0xD8)
				mga->tv_option = 0;
			if (bus->devices[i]->addr == 0xDA)
				mga->tv_option = 1;
		}
                break;
        }
}

static void _mgai2c_detach( struct i2c_bus* bus, int id )
{
        struct mga_dev* mga = (struct mga_dev*)bus->data;

        dprintk( "mgavideo: i2c device detach: %d\n", id );

        switch( id )
        {
        case I2C_DRIVERID_MSP3400:
                mga->hasaudio = 0;
                break;

        case I2C_DRIVERID_TUNER:
                mga->hastuner = 0;
                break;

        case I2C_DRIVERID_VIDEODECODER:
                mga->hasdecoder = 0;
                break;
        }
}

static void _mgai2c_setlines( struct i2c_bus* bus, int ctrl, int data )
{
        struct mga_dev* mga = (struct mga_dev*)bus->data;

        if( ctrl )
        {
        writeb( XGENIOCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) & ~SCL, mga->ctrl + X_DATAREG );
        }
        else
        {
        writeb( XGENIODATA, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) & ~SCL, mga->ctrl + X_DATAREG );
        writeb( XGENIOCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) | SCL, mga->ctrl + X_DATAREG );
        }

        if( data )
        {
        writeb( XGENIOCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) & ~SDA, mga->ctrl + X_DATAREG );
        }
        else
        {
        writeb( XGENIODATA, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) & ~SDA, mga->ctrl + X_DATAREG );
        writeb( XGENIOCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) | SDA, mga->ctrl + X_DATAREG );
        }

        udelay( I2C_DELAY );
}

static int _mgai2c_getdataline( struct i2c_bus* bus )
{
        struct mga_dev* mga = (struct mga_dev*)bus->data;

        writeb( XGENIOCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) & ~SDA, mga->ctrl + X_DATAREG );
        writeb( XGENIODATA, mga->ctrl + PALWTADD );
        return ( readb( mga->ctrl + X_DATAREG ) & SDA )?1:0;
}

/****************************************************************************
* private api : internal interface to the mga's video subsystem
****************************************************************************/
/******************************************************************************
 * This is the CODEC interface from mgadvd (as slightly modified as possible) *
 ******************************************************************************/
/* VERY UGLY CODE: Define aliases for the DVD variables */
#define mga_dev_s mga_dev

/* wait 'delay' microseconds */
int i34_wait(int delay)
{
	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout(delay * HZ/1000000);
	return 1;
}

/****************************************************************************
 * CODEC INTERFACE                                                          *
 ****************************************************************************/

/* This interface is made to work with the Zoran DVD Codec. More specifically it
   implements 8 bit VMI protocol A, with a big endian target chip.
   It uses a fixed buffer for bitstream transfer.
   The command/data buffers are prepared in central memory and sent to the
   framebuffer just at execution time.
   See the Matrox G{1,2,400} specs for details (p. 4-83 for G200)
*/

#define CODEC_BUFFER_SIZE	0x40000	/* 256K */
// #define CODEC_BUFFER_SIZE	0x20000	/* 128K */

/* CODEC registers */
#define CODECCTL        0x3e40
#define CODECMISCCTL	0x3e43
#define CODECADDR       0x3e44
#define CODECHOSTPTR    0x3e48
#define CODECHARDPTR    0x3e4c
#define CODECLCODE      0x3e50

#define PALWTADD        0x3c00
#define X_DATAREG       0x3c0a
#define XGENIOCTRL      0x2a
#define XGENIODATA      0x2b
#define VSTATUS		0x3e30
#define VICLEAR         0x3e34
#define VIEN            0x3e38
#define STATUS		0x1e14
#define IEN             0x1e1c

/*
 * CODECCTL fields and commands
 */
#define codecen 	0x0001	/* CODEC enable, invert to disable */
#define codecmode 	0x0002	/* CODEC I33 mode, invert for VMI mode */
#define cmdexectrig 	0x0004	/* command execution trigger, invert to prevent execution */
#define codecdatain 	0x0008	/* compression, invert for decompression */
#define vmimode 	0x0010	/* VMI mode B, invert for mode A */
#define stopcodec 	0x0020	/* stop after current field, invert to not stop */
#define codectransen 	0x0040	/* enable transfer, invert to disable */
#define codec_stalled 	(1<<12)	/* VSTATUS: current bitstream transfer is completed */


/************* START OF ZR36060 HANDLING ****************/

/* These 2 routines use a shadow register for CODECCTL, This is need
   because the CODECCTL register is a write only register.
*/

static void 
mgacodec_and_or_codecctl(struct mga_dev *mga, u16 and, u16 or)
{
	int	codecctl;

	codecctl = (mga->codecctl_shadow & and) | or;
	mga->codecctl_shadow = codecctl & ~0x04; /* alway remove execute bit */

	_mgawritew(mga, CODECCTL, codecctl);
}


static void 
mgacodec_and_or_codecmisc(struct mga_dev *mga, u16 and, u16 or)
{
	int	codecmisc;

	codecmisc = (mga->codecmisc_shadow & and) | or;
	mga->codecmisc_shadow = codecmisc;

	_mgawriteb(mga, CODECCTL+3, codecmisc);
}


static int
mgacodec_transfer_stall(struct mga_dev *mga)
{
	int	i;
	int	status;

	if (mga->codecctl_shadow & VC_TRANSEN) {
		printk("wait for transfer to stop\n");
		mgacodec_and_or_codecctl(mga, ~(VC_TRIG | VC_TRANSEN), 0);
	    
		for(i = 0, status = _mgaread(mga, VSTATUS);
			(status & VS_STALLED) == 0 && i++ < 1024;
			status = _mgaread(mga, VSTATUS));

		printk("transfer stopped status %x VS_STALLED %x i = %d\n", status, VS_STALLED, i);
	}

	return 0;
}

static int
mgacodec_execute_command(struct mga_dev *mga)
{
	int	i,tries;
	int	status;


	for(tries=1;tries < 6;tries++) {
	  /* tell the system where memory buffer for zr36060 exists */
	  _mgawrite(mga, CODECADDR, (mga->host_interface_base & ~0x03) | 0x01);

#ifdef		NEVER
	  mgacodec_transfer_stall(mga)
#endif
	  
	  /* remove any pending command complete */
	  _mgawrite(mga, VICLEAR, VI_CMDCMPL);

	  /* start codec command */
	  mgacodec_and_or_codecctl(mga, ~0xf0, VC_TRIG);

	  /* wait for codec command to complete */
	  for(i = 4096, status = _mgaread(mga, VSTATUS);
	      (status & VS_STALLED) == 0 && i--; 
	      status = _mgaread(mga, VSTATUS)) ;
	  
	  /* Ack the command complete interrupt */
	  _mgawriteb(mga, VICLEAR, VI_CMDCMPL);
	  
	  mga->codec_offset = 0;
	  
	  if (status & VS_CMDCMPL)
	    return 0;

	  mgacodec_and_or_codecctl(mga, ~VC_TRIG, 0);	

	  dprintk("Codec I/O Error #%d, restarting command\n",tries);
	}

	printk("mgavideo: no command complete on command 0x%x status 0x%x\n", 
	       readw(mga->command_base), status);

	return 1;
}

static	int
mgacodec_read_memory(struct zr36060_dev *zr36060, void *ptr, int mem, int size)
{
	int	last_read;
	int	last_exec;
	int	ind_1; /* direct ZR36060 registers */ /* bit 10-8 of param address */ 
	int	ind_2; /* bit 7-0 of param address */
	int	ind_3; /* DATA I/O */
	int	i;
	int	j;
	char	*buffer;
	struct	mga_dev *mga;

	mga = zr36060->private;
	buffer = ptr;

	if (mga->tv_option) { /* Marvel reg config */
		ind_1 = 0x0100;
		ind_2 = 0x0200;
		ind_3 = 3 << 10;
	} else 	/* This is the famous RRG register magic reg config */
	  {
	    ind_1 = 0x0600; /* everything treated with ^ 0x700 */
	    ind_2 = 0x0500;
	    ind_3 = 4 << 10;
	  }

	last_read = 0;
	last_exec = 0;

	for(i=0; i < size; i++, mem++) {
	  if (mga->codec_offset > (512 - 32)) { /* buffer full, read already now */
			if (i > 0) {
				writew(0x8201 | ind_3, 
				       mga->command_base + last_read);
			}

			if (mgacodec_execute_command(mga)) {
				printk("mgavideo: Codec read buffer error\n");
				return -1;
			}

			for(j = last_exec; j < i; j++) {
				*buffer++ = readb(mga->read_base + 
					 (j - last_exec) * 8);

			}

			last_exec = i;

		}
		writew(0x0002, mga->command_base + mga->codec_offset);
		mga->codec_offset += 8;

		/* write bits 8-10 of param address */ 
		writew((mem >> 8) | ind_1, 
			mga->command_base + mga->codec_offset);
		mga->codec_offset += 2;

		/* write bits 7-0 of param address */ 
		writew((mem & 0xff) | ind_2, 
			mga->command_base + mga->codec_offset);
		mga->codec_offset += 2;

		mga->codec_offset = (mga->codec_offset & ~7) + 8;

		/* queue a read from DATA I/O to read the addressed param memory */
		writew(0x8001 | ind_3, 
			mga->command_base + mga->codec_offset);
		last_read = mga->codec_offset;

		mga->codec_offset += 8;
	}

	if (last_read > 0) {
		writew(0x8201 | ind_3, mga->command_base + last_read);
	}

	if (mgacodec_execute_command(mga)) {
		printk("mgavideo: Codec read buffer error\n");
		return -1;
	}

	for(i=last_exec; i < size; i++) {
		*buffer++ = readb(mga->read_base + (i - last_exec) * 8);
	}

	return 0;
}

static int
mgacodec_write_memory(struct zr36060_dev *zr36060, void *mem, int start_reg, int length)
{
	int	offset;
	int	count;
	int	last_high;
	int	ind_1;
	int	ind_2;
	int	ind_3;
	char	*buffer;
	struct	mga_dev *mga;

	mga = zr36060->private;
	buffer = mem;

	if (mga->tv_option) {
		ind_1 = 0x0100;
		ind_2 = 0x0200;
		ind_3 = 0x0300;
	} else {
		ind_1 = 0x0600;
		ind_2 = 0x0500;
		ind_3 = 0x0400;
	}



	offset = (int)mga->command_base + 8;
	count = 0;
	last_high = -1;

	while(length--) {
		if (last_high != (start_reg >> 8)) {
			last_high = start_reg >> 8;
			writew((start_reg >> 8) | ind_1, offset);
			offset += 2;
			count++;
		}
		    
		writew((start_reg++ & 0xff) | ind_2, offset);
		writew((*buffer++ & 0xff) | ind_3, offset+2);
		offset+= 4;
		count += 2;

		if (count > 250) {
			writew(count | 0x200, mga->command_base);
			if (mgacodec_execute_command(mga)) {
				printk("mgavideo: Codec write buffer error\n");
				return -1;
			}
			offset = (int)mga->command_base + 8;
			count = 0;
		}
	}

	if (count) {
		writew(count | 0x200, mga->command_base);
		if (mgacodec_execute_command(mga)) {
			printk("mgavideo: Codec write buffer error\n");

			return -1;
		}
	}

	return 0;
}

static int
mgacodec_zr36060_overlay(struct zr36060_dev *zr36060, int flag)
{
	return 0;
}

static int
mgacodec_zr36060_set_jpg(struct zr36060_dev *zr36060, enum buz_codec_mode mode)
{
	return 0;
}

/* These functions control the MGA codec to interface with the JPEG codec */
static int
mgacodec_zr36060_transfer(struct zr36060_dev *zr36060, enum buz_codec_mode mode) 
{
	struct	mga_dev	*mga;
	void	*start;
	void	*start1;
	int	size;
	int	size1;
	int	count;
	int	ind_0;
	int	ind_1;
	int	ind_2;
	int	ind_3;

	mga = zr36060->private;

	if (mga->tv_option) {
		ind_0 = 0;
		ind_1 = 0x0100;
		ind_2 = 0x0200;
		ind_3 = 0x03 << 10;
	} else {
		ind_0 = 0x03;
		ind_1 = 0x0600;
		ind_2 = 0x0500;
		ind_3 = 0x04 << 10;
	}


	switch(mode) {

	case BUZ_MODE_MOTION_COMPRESS:
	case BUZ_MODE_STILL_COMPRESS:

		/* Enable both command complete and vsync interrupts */
		mgavideo_ivsync_enable(mga, 1);
		mgavideo_command_enable(mga, 1);

		/* Set the register to read fifo data from the zr36060 */
		_mgawriteb(mga, 0x3e41, ind_0);

		/* Lower the start line of the zr36060 */
		mgacodec_and_or_codecmisc(mga, ~0, 0x10);

		/* Read zr36060 read status (8) */
		writew(0x0002, mga->command_base + 0);
		writew(0x0000 | ind_1, mga->command_base + 8);
		writew(0x0008 | ind_2, mga->command_base + 10);
		writew(0x8001 | ind_3, mga->command_base + 16);

		/* Pause for compress and reread status (8) and stop */
		writew(0x0102, mga->command_base + 24);
		writew(0x0008 | ind_1 , mga->command_base + 32);
		writew(0x0000 | ind_2 , mga->command_base + 34);
		writew(0x8201 | ind_3, mga->command_base + 40);

		/* start up command and compress one field */
		mgacodec_and_or_codecctl(mga, ~0, 0x6f);

		/* raise the start line of the zr36060 */
		mgacodec_and_or_codecmisc(mga, ~0x10, 0x00);
		break;

	case BUZ_MODE_MOTION_DECOMPRESS:
	case BUZ_MODE_STILL_DECOMPRESS:
printk("sarting decompression\n");
		for(;;) {
			start = &mga->interface_base[mga->codec_tail];
			size = (258 * 1024) - mga->codec_tail;

			start1 = &mga->interface_base[mga->codec_head];
			size1 = mga->codec_head;

			count = zr36060_get_field(zr36060, start, size, 
				start1, size1);
			if (count == 0) {
				break;
			}
			mga->codec_tail += count;
			mga->codec_tail &= ((256 *1024) -1);
		}

		if (mga->codec_tail == mga->codec_head) {
printk("mga codec_tail %x = codec_head %x\n", mga->codec_tail, mga->codec_head);
			return 0;
		}

		/* Enable both command complete and vsync interrupts */
		mgavideo_ivsync_enable(mga, 1);
		mgavideo_command_enable(mga, 1);
		mgavideo_eoi_enable(mga, 1);


		/* Set the register to write fifo data to the zr36060 */
		_mgawriteb(mga, 0x3e41, ind_0);

		/* Lower the start line of the zr36060 */
		mgacodec_and_or_codecmisc(mga, ~0, 0x10);

		/* Read zr36060 read status (8) */
		writew(0x0002, mga->command_base + 0);
		writew(0x0000 | ind_1, mga->command_base + 8);
		writew(0x0008 | ind_2, mga->command_base + 10);
		writew(0x8001 | ind_3, mga->command_base + 16);

		/* Pasuse for compress and reread status (8) and stop */
		writew(0x0102, mga->command_base + 24);
		writew(0x0008 | ind_1 , mga->command_base + 32);
		writew(0x0000 | ind_2 , mga->command_base + 34);
		writew(0x8201 | ind_3, mga->command_base + 40);

		

printk("starting codec running\n");
		/* start up command and compress one field */
		mgacodec_and_or_codecctl(mga, ~0, 0x67);
		break;

	case BUZ_MODE_IDLE:
		mgacodec_transfer_stall(mga);
		break;
	}

	mga->mjpeg_codec_mode = mode;

	return 0;
}

static int
mgacodec_zr36060_set_video(struct zr36060_dev *zr36060, enum buz_codec_mode mode)
{
	int	temp;
	struct	mga_dev	*mga;

	mga = zr36060->private;

	switch(mode) {
	case BUZ_MODE_MOTION_COMPRESS:
	case BUZ_MODE_STILL_COMPRESS:
		temp = KS_OUTPUT_EXV;
		mgavideo_decoder(mga, KS0127_SET_OUTPUT, &temp);
		break;

	case BUZ_MODE_MOTION_DECOMPRESS:
	case BUZ_MODE_STILL_DECOMPRESS:
		temp = KS_INPUT_YUV656_60HZ;
		mgavideo_decoder(mga, KS0127_SET_INPUT, &temp);
		temp = KS_OUTPUT_YUV656E;
		mgavideo_decoder(mga, KS0127_SET_OUTPUT, &temp);
		break;

	case BUZ_MODE_IDLE:
		temp = KS_OUTPUT_YUV656E;
		mgavideo_decoder(mga, KS0127_SET_OUTPUT, &temp);
		break;
	}
	return 0;
}

static int
mgacodec_zr36060_wakeup(struct zr36060_dev *zr36060, int flag) 
{
	struct	mga_dev	*mga;

	mga = zr36060->private;

	dprintk("mgacodec_zr36060_wakeup: %s\n", 
		flag ? "Making active" : "Putting to sleep");

	/* reset the codec interface engine */
	mgacodec_and_or_codecctl(mga, 0x00, 0x00);
	mgacodec_and_or_codecctl(mga, 0x00, 0x03);
	mgacodec_and_or_codecctl(mga, 0x00, 0x03);

	/* pulse the reset line */
	mgacodec_and_or_codecmisc(mga, ~0x0f, 0x07);
	mdelay(10);
	mgacodec_and_or_codecmisc(mga, ~0x0f, 0x03);
	mdelay(1);
	mgacodec_and_or_codecmisc(mga, ~0x0f, 0x07);

	/* only need to reset chip. Do not wake it up yet. */
	if (flag == 0)
		return 0;

	/* pulse the sleep line */
	mgacodec_and_or_codecmisc(mga, ~0x0f, 0x07);
	mdelay(100);
	mgacodec_and_or_codecmisc(mga, ~0x0f, 0x0f);
	mdelay(1);
	mgacodec_and_or_codecmisc(mga, ~0x0f, 0x07);
	mdelay(200);

	/* tell the system where memory buffer for zr36060 exists */
	_mgawrite(mga, CODECADDR, (mga->host_interface_base & ~0x03) | 0x01);

	return 0;
}

/******************** END OF ZR36060 ************************************/

/****************************************************************************
* mgavideo api : public interface to the mga's video subsystem
****************************************************************************/
/* component device control */
void mgavideo_tuner( struct mga_dev* mga, unsigned int cmd, void* arg )
{
        if( mga == NULL )
                return;

        if( mga->hastuner )
                i2c_control_device( &(mga->i2c), I2C_DRIVERID_TUNER, cmd, arg );
}

void mgavideo_audio( struct mga_dev* mga, unsigned int cmd, void* arg )
{
        if( mga == NULL )
                return;

        if( mga->hasaudio )
                i2c_control_device( &(mga->i2c), I2C_DRIVERID_MSP3400, cmd, arg );

}

void mgavideo_decoder( struct mga_dev* mga, unsigned int cmd, void* arg )
{
        if( mga == NULL )
                return;

	/* copy these important values also on the mga level */
	if (cmd == KS0127_SET_INPUT) 
	  mga->zr36060.params.input = * (int *)arg;
	if (cmd == KS0127_SET_STANDARD)
	  mga->zr36060.params.norm = * (int *)arg;

	  switch(mga->zr36060.params.norm) {

	  case KS_STD_PAL:
		mga->video_activelines = _VINHGT_PAL;
		break;

	  case KS_STD_NTSC:
		mga->video_activelines = _VINHGT_NTSC;
		break;
	  }

        if( mga->hasdecoder )
                i2c_control_device( &(mga->i2c), I2C_DRIVERID_VIDEODECODER, cmd, arg );
}

/* These are the functions for the ZORAN<->MGA codec interface */

int mgavideo_zr36060_attach(struct mga_dev* mga)
{
  mga->zr36060.params.major_version = MAJOR_VERSION;
  mga->zr36060.params.minor_version = MINOR_VERSION;

  /* Don't know how to handle : */
  /* default_norm  ? 1 : 0;  Avoid nonsense settings from user */
  mga->zr36060.params.input    = KS_INPUT_TUNER;
  mga->zr36060.params.norm     = KS_STD_PAL;
  /* default_input ? 1 : 0;  Avoid nonsense settings from user */ 

  zr36060_attach(&mga->zr36060);

  return 0;
}

int mgavideo_querycaps( struct mga_dev* mga )
{
        int caps = 0;

        if( mga == NULL )
                return 0;

        if( mga->hastuner )
                caps |= MGAVIDEO_HAS_TUNER;

        if( mga->hasaudio )
                caps |= MGAVIDEO_HAS_AUDIO;

        if( mga->hasdecoder )
                caps |= MGAVIDEO_HAS_DECODER;

        if( mga->hasmjpeg )
                caps |= MGAVIDEO_HAS_MJPEG;

        if( mga->hastvout )
                caps |= MGAVIDEO_HAS_TVOUT;

        return caps;
}

#ifdef	NEVER
void 
mgavideo_lock_video(struct mga_dev* mga, int *stamp, int *field, u8 **buffer)
{
	stamp[0] = stamp[1] = -1;
	field[0] = field[1] = 0;
	buffer[0] = buffer[1] = 0;

	if (mga->field_req == FIELD_INTERLACED) {
		if (mga->buffer_stamp[0] && mga->buffer_stamp[1]) {
			mga->buffer_status[0] |= BUFFER_LOCKED;
			mga->buffer_status[1] |= BUFFER_LOCKED;
			stamp[0] = mga->buffer_stamp[0];
			stamp[1] = mga->buffer_stamp[1];
			field[0] = mga->buffer_status[0];
			field[1] = mga->buffer_status[1];
			buffer[0] = (u8 *)mga->fb + mga->vin_addr[0];
			buffer[1] = (u8 *)mga->fb + mga->vin_addr[1];
			return;
		}

		if (mga->buffer_stamp[2] && mga->buffer_stamp[3]) {
			mga->buffer_status[0] |= BUFFER_LOCKED;
			mga->buffer_status[1] |= BUFFER_LOCKED;
			stamp[0] = mga->buffer_stamp[0];
			stamp[1] = mga->buffer_stamp[1];
			field[0] = mga->buffer_status[0];
			field[1] = mga->buffer_status[1];
			buffer[0] = (u8 *)mga->fb + mga->vin_addr[0];
			buffer[1] = (u8 *)mga->fb + mga->vin_addr[1];
			return;
		}
	} else if (mga->field_req == FIELD_1) {
		if (mga->buffer_stamp[0] &&  mga->buffer_stamp[2]) {
			if (mga->buffer_stamp[0] < mga->buffer_stamp[2])  {
				/* buffer 0 has wanted field */
				mga->buffer_status[0] |= BUFFER_LOCKED;
				*stamp = mga->buffer_stamp[0];
				*field = mga->buffer_status[0];
				buffer[0] = (u8 *)mga->fb + mga->vin_addr[0];
				return;
			} else {
				/* buffer 2 has wanted field */
				mga->buffer_status[2] |= BUFFER_LOCKED;
				*stamp = mga->buffer_stamp[2];
				*field = mga->buffer_status[2];
				buffer[0] = (u8 *)mga->fb + mga->vin_addr[2];
				return;
			}
		} else if (mga->buffer_stamp[0]) {
			/* buffer 0 has wanted field */
			mga->buffer_status[0] |= BUFFER_LOCKED;
			*stamp = mga->buffer_stamp[0];
			*field = mga->buffer_status[0];
			buffer[0] = (u8 *)mga->fb + mga->vin_addr[0];
			return;
		} else if (mga->buffer_stamp[2]) {
			/* buffer 3 has wanted field */
			mga->buffer_status[2] |= BUFFER_LOCKED;
			*stamp = mga->buffer_stamp[2];
			*field = mga->buffer_status[2];
			buffer[0] = (u8 *)mga->fb + mga->vin_addr[2];
			return;
		}
	} else if (mga->field_req == FIELD_2) {
		if (mga->buffer_stamp[1] &&  mga->buffer_stamp[3]) {
			if (mga->buffer_stamp[1] < mga->buffer_stamp[3])  {
				/* buffer 1 has wanted field */
				mga->buffer_status[1] |= BUFFER_LOCKED;
				*stamp = mga->buffer_stamp[1];
				*field = mga->buffer_status[1];
				buffer[0] = (u8 *)mga->fb + mga->vin_addr[1];
				return;
			} else {
				/* buffer 3 has wanted field */
				*stamp = mga->buffer_stamp[3];
				*field = mga->buffer_status[3];
				mga->buffer_status[3] |= BUFFER_LOCKED;
				buffer[0] = (u8 *)mga->fb + mga->vin_addr[3];
				return;
			}
		} else if (mga->buffer_stamp[1]) {
			/* buffer 0 has wanted field */
			mga->buffer_status[1] |= BUFFER_LOCKED;
			*stamp = mga->buffer_stamp[1];
			*field = mga->buffer_status[1];
			buffer[0] = (u8 *)mga->fb + mga->vin_addr[1];
			return;
		} else if (mga->buffer_stamp[3]) {
			/* buffer 3 has wanted field */
			mga->buffer_status[3] |= BUFFER_LOCKED;
			*stamp = mga->buffer_stamp[3];
			*field = mga->buffer_status[3];
			buffer[0] = (u8 *)mga->fb + mga->vin_addr[3];
			return;
		}
	}

	return;
}

void 
mgavideo_unlock_video(struct mga_dev* mga, int *stamp, int flag)
{
	int	i;
	int	j;

	for(i=0; i < 2; i++) {
		for(j=0; j < 4; j++) {
			if (mga->buffer_stamp[j] == stamp[i]) {
				if (flag) {
					mga->buffer_stamp[j] = 0;
					mga->buffer_status[j] = 0;
				} else {
					mga->buffer_status[j] &= ~BUFFER_LOCKED;
				}
				break;
			}
		}
	}
}
#endif

static void _detect_tvt( struct mga_dev* mga )
{
        int msp, tuner_addr, tuner;

        if( mga->hastuner ) {
                mgavideo_tuner( mga, TUNER_GET_ADDRESS, &tuner_addr );

                if( mga->hasaudio ) {
                        mgavideo_audio( mga, AUDC_GET_UNIT, &msp );
                }

                tuner = TUNER_ABSENT;

                switch( tuner_addr ) {
                case 0xc0:
                        /* Samsung MULTI PAL - unsupported by tuner.c? */
                        break;

                case 0xc2:
                        tuner = TUNER_PHILIPS_NTSC;

                        if( (msp&0xff) == 0x07 ) { /* Usa audio */
                                tuner = TUNER_PHILIPS_NTSC;
                        } else {
                        /* Samsung SECAM - unsupported by tuner.c? */
                        }
                        break;
                
                case 0xc4:
                        tuner = TUNER_PHILIPS_PAL_I;
                        break;

                case 0xc6:
                        tuner = TUNER_PHILIPS_SECAM;

                        if( (msp&0xff) == 0x07 ) { /* Usa audio */
                        /* Samsung NTSC - unsupported by tuner.c? */
                        } else {
			  tuner = TUNER_PHILIPS_SECAM; /* RRG, PAL version */
                        }
                }
                
                mgavideo_tuner( mga, TUNER_SET_TYPE, &tuner );
        }
}

/****************************************************************************
* interrupt processing
****************************************************************************/
int mgavideo_codec_size(int start, int end)
{
	if (start < end) {
		return (end - start) * 4;
	} 

	return ((256 * 1024 - start) + end) * 4;
}

static void 
mgavideo_vsync(struct mga_dev *mga)
{
	int	field_buffer_match;
	int	i;
	int	program_window;

	/* clear sync interrupt */
	_mgawrite( mga, VICLEAR, VI_IVSYNC );

	/* Update the preview window */
        if (mga->preview_active) {
		if (! mga->interlaced_preview) {
			/* Non interlaced preview : display the fields separately */
			/* with interpolation according to the field type. */
	        	if (!(mga->status & VS_FIELD)) {
        			mga->besctl_value = (mga->besctl_value & ~(3<<25));
        			_mgawrite(mga, BESA1ORG,
        				mga->vin_addr[mga->active_buffer]
        				+ mga->bes_origin_topfield);
   			} else {
        			mga->besctl_value = (mga->besctl_value & ~(3<<25)) | (1<<25);
        			_mgawrite(mga, BESA2ORG,
        				mga->vin_addr[mga->active_buffer]
        				+ mga->bes_origin_bottomfield);
   			}
       			_mgawrite(mga, BESCTL, mga->besctl_value);
        	} else {
			/* If interlaced preview, and the field is the last of an */
        		/* image, display the image built with the preceeding field. */
        		if ((mga->status & VS_FIELD) != VS_FIRSTFIELD) {
	        		mga->besctl_value = (mga->besctl_value & ~(3<<25));
        			_mgawrite(mga, BESA1ORG,
        				mga->vin_addr[mga->active_buffer & (~1)]
        				+ mga->bes_origin_topfield);
        		}
       			_mgawrite(mga, BESCTL, mga->besctl_value);
        	}
        }

        /* Update data for the buffer which has been grabbed */
        mga->buffer_flags[mga->active_buffer] &= ~(BUFFLAG_SERVERLOCK + BUFFLAG_FIELDTYPE);
        if (mga->status & VS_FIELD)
	        mga->buffer_flags[mga->active_buffer] |= BUFFLAG_FIELDTYPE;
	mga->grabbed++;
	mga->buffer_stamp[mga->active_buffer] = mga->grabbed;
	
        /* Test if the fields come in the correct buffers (for interlaced preview) */
       	field_buffer_match = ( ((mga->status & VS_FIELD)?1:0) == (mga->active_buffer & 1) );

        /* The buffer being grabbed now has been chosen during the last vsync interrupt. */
        mga->active_buffer = mga->next_active_buffer;
        mga->next_active_buffer = -1;

        /* Find which buffer to use next. Rules depend on the preview mode. */
        /* Grabbing is easier in non-interlaced preview mode (less constraints) */
	if (mga->interlaced_preview) {
	        /* Choose the next buffer to receive data. */
        	/* We do our best to group the fields and build complete images. */
		/* Default choice, use the circular order. */
		mga->next_active_buffer = (mga->active_buffer + 1) & 3;
		/* If the field-type doesn't match with the buffer number "oddity" then take another */
		/* one. Do the test for first fields only (the next field it being grabbed and is */
		/* also wrong, so don't exchange next time). */
		if ((mga->status & VS_FIELD) == VS_FIRSTFIELD)
			if (! field_buffer_match)
				mga->next_active_buffer = (mga->active_buffer + 2) & 3;
		/* If this one is locked, preview window sacrificed : use the normal criteria. */
		/* We could be smarter, but with higher risk of dropped frames in memory grabbing */
		if (mga->buffer_flags[mga->next_active_buffer] & (BUFFLAG_CLIENTLOCK+BUFFLAG_SERVERLOCK))
			mga->next_active_buffer = -1;
	}
	if (mga->next_active_buffer == -1) {
		/* Best grabbing politics applied (keep wanted buffers as long as possible) : */
		/* First choice : oldest unwanted buffer */
		/* Second choice : oldest unlocked buffer */
		int unwanted_found = 0;
		u32 oldeststamp = 0xffff;
		mga->next_active_buffer = 0;
		for (i = 0; i < 4; i++) {
			if ((mga->buffer_flags[i] & (BUFFLAG_CLIENTLOCK | BUFFLAG_SERVERLOCK)) == 0) {
				int field_type, wanted;
				field_type =
					((mga->buffer_flags[i] & BUFFLAG_FIELDTYPE)?FIELD_2:FIELD_1);
				wanted = field_type & mga->field_req;
				if (wanted) {
					if (!unwanted_found && oldeststamp > mga->buffer_stamp[i]) {
						oldeststamp = mga->buffer_stamp[i];
						mga->next_active_buffer = i;
					}
				} else if (!unwanted_found || oldeststamp > mga->buffer_stamp[i]) {
					oldeststamp = mga->buffer_stamp[i];
					mga->next_active_buffer = i;
					unwanted_found = 1;
				}
			}
		}
	}
	/* Lock the chosen buffer */
        mga->buffer_flags[mga->next_active_buffer] |= BUFFLAG_SERVERLOCK;

        /* Program the VIN windows to point to the chosen buffer */
        /* On the g200, we remember which window we programmed last time. */
        /* On the g400, we cannot predict in which window the data will be put */
        /* (it depends on the field type). So we make all windows point to the */
        /* same buffer. Here X1 mode is assumed. */
        if (mga->mga_type == MGA_G200) {
		mga->vin_active = (mga->vin_active+1) & 1;
	       	_mgawrite(mga, VINNEXTWIN, mga->vin_active);
	       	program_window = (mga->vin_active == 0) ? 1 : 2;
	} else {
		program_window = 3;
	}
	if (program_window & 1) {
		_mgawrite(mga, VINADDR0, mga->vin_addr[mga->next_active_buffer]);
		_mgawrite(mga, VBIADDR0, mga->vbi_addr[0]);
		_mgawrite(mga, VINCTL0,  mga->vinctl_value);
	}
	if (program_window & 2) {
		_mgawrite(mga, VINADDR1, mga->vin_addr[mga->next_active_buffer]);
		_mgawrite(mga, VBIADDR1, mga->vbi_addr[1]);
		_mgawrite(mga, VINCTL1,  mga->vinctl_value);
	}

return;
#ifdef	NEVER
	/* clear sync interrupt */
	_mgawrite( mga, VICLEAR, VI_IVSYNC );

	/* Preview the current buffer */
	/* We use the BES window which has been programmed for the current field type */
	/* (if the field is a top field, use window 1, otherwise, use window 2) */
        if (mga->preview_active) {
        	if (!(mga->status & VS_FIELD)) {
        		mga->besctl_value = (mga->besctl_value & ~(3<<25));
        		_mgawrite(mga, BESA1ORG,
        			mga->vin_addr[mga->buffer_active]+mga->bes_origin_topfield);
        	} else {
        		mga->besctl_value = (mga->besctl_value & ~(3<<25)) | (1<<25);
        		_mgawrite(mga, BESA2ORG,
        			mga->vin_addr[mga->buffer_active]+mga->bes_origin_bottomfield);
        	}
        	_mgawrite(mga, BESCTL, mga->besctl_value);
        }

        mga->buffer_active = (mga->buffer_active + 1) & 3;

        /* Select the next buffer to receive video data */
        /* On the g200, we remember which window we programmed last time. */
        /* On the g400, we cannot predict in which window the data will be put */
        /* (it depends on the field type). So we make all windows point to the */
        /* same buffer. Here X1 mode is assumed. */
        if (mga->mga_type == MGA_G200) {
		mga->vin_active = (mga->vin_active+1) & 1;
	       	_mgawrite(mga, VINNEXTWIN, mga->vin_active);
	       	_mgawrite(mga,
	       		(mga->vin_active == 0) ? VINADDR0 : VINADDR1,
	       		mga->vin_addr[(mga->buffer_active+1)&3]);
	} else if (mga->mga_type == MGA_G400) {
	       	_mgawrite(mga, VINADDR0, mga->vin_addr[(mga->buffer_active+1)&3]);
	       	_mgawrite(mga, VINADDR1, mga->vin_addr[(mga->buffer_active+1)&3]);	
	}
	/* If G400, return for now */
	if (mga->mga_type == MGA_G400) 
          return;  

#ifdef	NEVER
return;

	/* If we do not want interrupt then ignore them */
	if ((mga->interrupt_mask & VS_IVSYNC) == 0) {
		return;
	}

printk("\n\nmgavideo: status = %x head = %x tail = %x queue %x %x %x %x\n",
mga->status, head, tail, queue[0], queue[1], queue[2], queue[3]);

	if (mga->status & VS_VIDEO) {
		mga->buffer_status[queue[head]] = field[head];
		mga->buffer_stamp[queue[head]] = mga->grabbed++;
		head++;
		head &= 3;
	}

	if ((mga->status & VS_FIELD) == 0) {
	  mga->frame_next = 
	    (mga->frame_next == FRAME_1) ? FRAME_2 : FRAME_1;
	  
	  if ((mga->buffer_status[mga->frame_next == FRAME_1 ? 0 : 2] & BUFFER_LOCKED) ||
	      (mga->buffer_status[mga->frame_next == FRAME_1 ? 1 : 3] & BUFFER_LOCKED)) {
	    mga->frame_next = 
	      (mga->frame_next == FRAME_1) ? FRAME_2 : FRAME_1;
	  }
	  
		if (mga->status & VS_FIELD) {
		  mga->buffer_stamp[0] = 0;
		  mga->buffer_status[0] = 0;
		  
		  mga->buffer_stamp[1] = 0;
		  mga->buffer_status[1] = 0;
		  memset(mga->vin_addr[1]+mga->fb, 0, 32767 * 2);
		  memset(mga->vin_addr[0]+mga->fb, 0, 32767 * 2);
		} else {
		  mga->buffer_stamp[2] = 0;
		  mga->buffer_status[2] = 0;
		  
		  mga->buffer_status[3] = 0;
		  mga->buffer_stamp[3] = 0;
		  memset(mga->vin_addr[3]+mga->fb, 0, 32767 * 2);
		  memset(mga->vin_addr[2]+mga->fb, 0, 32767 * 2);
		}
	}
	
	buffer_next = (((mga->status & VS_FIELD) ? 1 : 0)) |
	  ((mga->frame_next == FRAME_1) ? 0 : 2);

#ifdef	NEVER	
	printk("mgavideo: status: %x field_req %x field_used %x(%d) %x(%d) %x(%d) %x(%d) next buffer %d\n",
	       mga->status, mga->field_req, 
	       mga->buffer_status[0], mga->buffer_stamp[0],
	       mga->buffer_status[1], mga->buffer_stamp[1],
	       mga->buffer_status[2], mga->buffer_stamp[2],
	       mga->buffer_status[3], mga->buffer_stamp[3],
	       buffer_next);
#endif

	if ((mga->field_req & FIELD_INTERLACED) || 
	    (mga->field_req & ((buffer_next & 1) ?  FIELD_2 : FIELD_1))) {
#ifdef	NEVER
	  printk("mgavideo: wanted next field %d req buffer_next %d stamps %d %d %d %d\n", 
		 mga->field_req, buffer_next,
		 mga->buffer_stamp[0],
		 mga->buffer_stamp[1],
		 mga->buffer_stamp[2],
		 mga->buffer_stamp[3]);
#endif
	  if (mga->buffer_stamp[buffer_next] == 0) {
	    if (!(buffer_next & 1)) {
#ifdef	NEVER
	      printk("mgavideo: vin_addr0: buffer_next %d\n", buffer_next);
#endif
	      queue[tail] = buffer_next;
	      field[tail] = 1;
	      tail++;
	      tail &= 3;
	      _mgawrite(mga, VINADDR0, 
			mga->vin_addr[buffer_next]);
	      _mgawrite(mga, VINNEXTWIN, 0);
	    } else {
	      queue[tail] = buffer_next;
	      field[tail] = 2;
	      tail++;
	      tail &= 3;
#ifdef	NEVER
	      printk("mgavideo: vin_addr1: buffer_next %d\n", buffer_next);
#endif
	      _mgawrite(mga, VINNEXTWIN, 1);
	      _mgawrite(mga, VINADDR1, 
			mga->vin_addr[buffer_next]);
	    }
	  }
	}

	/* queue bottom half if present, and requested */
#ifdef	NEVER
	if(mga->tqnode_dpc.routine && mga->interrupt_lock == 0) {
	  if (mga->field_req & FIELD_INTERLACED) {
	    if ((mga->buffer_stamp[0] && mga->buffer_stamp[1]) ||
		(mga->buffer_stamp[2] && mga->buffer_stamp[3])) {
	      mga->interrupt_lock++;
	      queue_task(&mga->tqnode_dpc, &tq_immediate);
	      mark_bh(IMMEDIATE_BH);
	    }
	  } else if (mga->buffer_stamp[0] || mga->buffer_stamp[1] ||
		     mga->buffer_stamp[2] || mga->buffer_stamp[3]) {
	    mga->interrupt_lock++;
	    queue_task(&mga->tqnode_dpc, &tq_immediate);
	    mark_bh(IMMEDIATE_BH);
	  }
	}

#else
	if (mga->interrupt_lock == 0) {
		mga->interrupt_lock++;
		queue_task(&mga->tqnode_dpc, &tq_immediate);
		mark_bh(IMMEDIATE_BH);
	}
#endif
#endif
#endif
}

static void interrupt_hw(int irq, void *v, struct pt_regs *regs)
{
        struct mga_dev* mga = (struct mga_dev*)v;
	char	*start;
	char	*start1;
	int	size;
	int	size1;
	int	count;
	
        if (mga == NULL)
                return;

        /* read status register */
        mga->status = _mgaread( mga, VSTATUS );

	if (mga->status & VS_IVSYNC)  {
		mgavideo_vsync(mga);
	}
	if (mga->status & VS_CMDCMPL) 
	  {
		printk("mgavideo: CTE command complete interrupt\n");
		_mgawrite( mga, VICLEAR, VI_CMDCMPL);

		mgacodec_and_or_codecctl(mga, 0x00, 3); /* reset the codec */
		udelay(1);

		if (mga->mjpeg_codec_mode == BUZ_MODE_MOTION_COMPRESS) {
		  int new_lcode; 

printk("zr36060: status %x %x\n", readb(mga->read_base), readb(mga->read_base+8));

		/* retrieve buffer end position of the read JPEG */
		new_lcode = _mgaread(mga, CODECLCODE); 
		mgacodec_zr36060_transfer(&mga->zr36060, 
			mga->mjpeg_codec_mode);

		  if (new_lcode != mga->lcode) {
		    if (mga->lcode < new_lcode) {
		      zr36060_new_field(&mga->zr36060, 
				mga->interface_base + mga->lcode * 4,
				(new_lcode - mga->lcode) * 4, NULL, 0);
		    } else {
		      zr36060_new_field(&mga->zr36060, 
				mga->interface_base + mga->lcode * 4,
				(256 * 1024 - (mga->lcode * 4)),
				mga->interface_base, new_lcode * 4);
		    }
		    
		    mga->lcode = new_lcode;
		  }
		}

		
	  }
	if (mga->status & VS_BLVL) 
	  {
	    /* good for refilling the queue in decompression */
	    /* and emptying it in compression */
#ifdef	NEVER
	    printk("mgavideo: CTE buffer level interrupt\n");
#endif
	    _mgawrite( mga, VICLEAR, VI_BLVL);

	  }
	if (mga->status & VS_EOI)  
	  {
int new_lcode;
	    /* good for flushing the decompression stream. */
	    /* and for syncing on a special image in decompression */
	    /* and for advancing the stat mode pointer in compression */
	    printk("mgavideo: CTE decompression EOI (= end of image) interrupt\n");
	    _mgawrite( mga, VICLEAR, VI_EOI);
new_lcode = _mgaread(mga, CODECLCODE); 
printk("codec lcode = %d\n", new_lcode);
 	    mgacodec_zr36060_transfer(&mga->zr36060, mga->mjpeg_codec_mode);
            zr36060_done(&mga->zr36060);
	  }
	if (mga->status & VS_STALLED) 
	  {
	    /* This is an underflow condition in compression */
	    printk("mgavideo: CTE Stalled interrupt\n");
	    _mgawrite(mga, VICLEAR, VI_STALLED);
	  }
}
/* enable interrupts */

void 
mgavideo_command_enable(struct mga_dev* mga, int en)
{
        if (mga == NULL)
                return;

        dprintk("mgavideo: %s command interrupt\n", en ? "enable" :" disable");

        _mgawrite(mga, VICLEAR, VI_CMDCMPL);

	if (en)
		mga->interrupt_mask |= VI_CMDCMPL;
	else
		mga->interrupt_mask &= ~VI_CMDCMPL;

	dprintk("mgavideo: interrupt_mask: %x\n", mga->interrupt_mask);
        _mgawrite(mga, VIEN, mga->interrupt_mask);
}



void 
mgavideo_buffer_enable(struct mga_dev* mga, int en)
{
        if (mga == NULL)
                return;

        dprintk("mgavideo: %s buffer level interrupt\n", en ? "enable" :" disable");

        _mgawrite(mga, VICLEAR, VI_BLVL);

	if (en)
		mga->interrupt_mask |= VI_BLVL;
	else
		mga->interrupt_mask &= ~VI_BLVL;

	dprintk("mgavideo: interrupt_mask: %x\n", mga->interrupt_mask);
        _mgawrite(mga, VIEN, mga->interrupt_mask);
}




void 
mgavideo_eoi_enable(struct mga_dev* mga, int en)
{
        if (mga == NULL)
                return;

        dprintk("mgavideo: %s eoi interrupt\n", en ? "enable" :" disable");

        _mgawrite(mga, VICLEAR, VI_EOI);

	if (en)
		mga->interrupt_mask |= VI_EOI;
	else
		mga->interrupt_mask &= ~VI_EOI;

	dprintk("mgavideo: interrupt_mask: %x\n", mga->interrupt_mask);
        _mgawrite(mga, VIEN, mga->interrupt_mask);
}



/* enable interrupts */
void 
mgavideo_ivsync_enable(struct mga_dev* mga, int en)
{

        if (mga == NULL)
                return;

        dprintk("mgavideo: %s interrupt\n", en ? "enable" :" disable");

        _mgawrite(mga, VICLEAR, VI_IVSYNC);

	if (en)
		mga->interrupt_mask |= VI_IVSYNC;
	else
		mga->interrupt_mask &= ~VI_IVSYNC;

	dprintk("mgavideo: interrupt_mask: %x\n", mga->interrupt_mask);
        _mgawrite(mga, VIEN, mga->interrupt_mask);
}


/* register a routine to be called for each frame */
void mgavideo_register_bh( struct mga_dev* mga, void(*routine)(void*), void* data )
{
        mga->tqnode_dpc.next = NULL;
        mga->tqnode_dpc.sync = 0;
        mga->tqnode_dpc.routine = routine;
        mga->tqnode_dpc.data = data;
}

/*
 * reset the the hardware for dma cap for new size. Also need to
 * update the ks0127 to scale for new size.
 */
void  mgavideo_set_dims(struct mga_dev *mga, int width, int height, int flags)
{
	mgavideo_decoder(mga, KS0127_SET_WIDTH, &width);
	mgavideo_decoder(mga, KS0127_SET_HEIGHT, &height);

	mga->field_req = flags;

	memset(mga->buffer_stamp, 0, sizeof(mga->buffer_stamp));
	memset(mga->buffer_status, 0, sizeof(mga->buffer_status));
	mga->vinpitch = width;

	if (mga->mga_type == MGA_G400) {
        	_mgawrite(mga, VINCTL0, ( VIN_ON | (mga->vinpitch >> 3) << 3));
        	_mgawrite(mga, VINCTL1, ( VIN_ON | (mga->vinpitch >> 3) << 3));
	} else {
                _mgawrite(mga, VINCTL0, ( VIN_ON | (mga->vinpitch >> 2) << 3));
                _mgawrite(mga, VINCTL1, ( VIN_ON | (mga->vinpitch >> 2) << 3));
	}
}

/****************************************************************************
* back-end scaler
****************************************************************************/
void* mgavideo_get_base( struct mga_dev* mga )
{
        return mga->fb;
}

/*
New BES code (Benoit Mathieu)

*/

/* Parameters for the preview window */
#define BES_HFILTER_ON 1
#define BES_VFILTER_ON 1
#define BES_ACC2XZOOM_ON  0
#define BES_DITHER_ON 1
#define BES_BW_ON 0

void mgavideo_set_window(struct mga_dev *mga, int x, int y, int w, int h)
{
	u32 cropped_width, cropped_height;
	u32 offset_left, offset_right, offset_top, offset_bottom;
	/* Values for the BES */
	u32 beshsrclst, intrep, ifactor, ifactorbetter, condition;
	u32 roundoff, acczoom, beshiscal, beshsrcst, beshsrcend, besviscal;
	u32 subpixelcomp, dataformat, origin_adress1, origin_adress2;
	u32 weight1, weight1s, weight2, weight2s, tmp, absval;
	u32 besv1srclst, besv2srclst, besv1srcstp, besv2srcstp;
	u32 bes_pitch;
	u32 win_height;
	
	if (mga->interlaced_preview) {
		win_height = 2 * mga->video_activelines;
		bes_pitch = _VINWID;
	} else {
		win_height = mga->video_activelines;
		bes_pitch = 2 * _VINWID;
	}
	
	/* Preparing parameters for the BES */
	/* This is and almost exact implementation of the G200 specs */
	/* (horizontal mirroring skipped for the moment) */

	/* Check if the window is on the desktop */
	if (w < 2) w = 2;
	if (w > 16383) w = 16383;
	if (h < 2) h = 2;
	if (h > 16383) h = 16383;
 	if (x >= (signed int) mga->desktop_width || x + w <= 0
 	    || y >= (signed int) mga->desktop_height || y + h <= 0) {

	    /* Window is not on the desktop -> no preview */
	    mga->besctl_value = 0;
	    _mgawrite(mga, BESCTL, 0);
	    return;
	}
	
	/* Horizontal cropping */
	cropped_height = win_height - mga->bes_crop_top - mga->bes_crop_bottom;
	cropped_width = _VINWID - mga->bes_crop_left - mga->bes_crop_right;
	offset_left = (x < 0) ? -x : 0;
	offset_right = (x+w > mga->desktop_width) ? x+w-mga->desktop_width : 0;
	beshsrclst = _VINWID - 1;

	/* Inverse Horiz scaling factor */
	if (BES_HFILTER_ON)
		intrep = (w == cropped_width) ? 0 : 1;
	else
		intrep = (w >= cropped_width) ? 0 : 1;
	ifactor = ((cropped_width - intrep) << 14) / (w - intrep);
	ifactorbetter = ((cropped_width - intrep) << 20) / (w - intrep) + intrep;
	condition = ( ((ifactorbetter * (w-1)) >> 20) > ((ifactor * (w-1)) >> 14) );
	roundoff = (condition && !BES_HFILTER_ON) ? 1 : 0;
	acczoom = (BES_ACC2XZOOM_ON) ? 2 : 1;
	beshiscal = (acczoom * ifactor) + roundoff;
	if (beshiscal >= (32 << 14))
		beshiscal = (32 << 14) - 1;
	/* Lower bound for beshiscal is already tested (w < 16384) */
	/* Horizontal source positioning */
	beshsrcst = (mga->bes_crop_left << 14) + offset_left * (ifactor + roundoff);
	beshsrcend = beshsrcst + ((w - offset_left - offset_right - 1) / acczoom)
	             * (acczoom * ifactor + roundoff);

	/* Vertical cropping */
	cropped_height = win_height - mga->bes_crop_top - mga->bes_crop_bottom;
	offset_top = (y < 0) ? -y : 0;
	offset_bottom = (y+h > mga->desktop_height) ? y+h-mga->desktop_height : 0;
	
	/* Inverse Vertical scaling factor */
	if (BES_VFILTER_ON)
		intrep = (w == cropped_height) ? 0 : 1;
	else
		intrep = (w >= cropped_height) ? 0 : 1;
	ifactor = ((cropped_height - intrep) << 14) / (h - intrep);
	ifactorbetter = ((cropped_width - intrep) << 20) / (h - intrep) + intrep;
	condition = ( ((ifactorbetter * (h-1)) >> 20) > ((ifactor * (h-1)) >> 14) );
	roundoff = (condition && !BES_VFILTER_ON) ? 1 : 0;
	besviscal = ifactor + roundoff;
	if (besviscal >= (32 << 14))
		besviscal = (32 << 14) - 1;
	/* Vertical subpixel compensation */
	if (besviscal >= (1 << 14)) /* Downscaling */
		subpixelcomp = besviscal - (1 << 13);
	else                        /* Upscaling */
		subpixelcomp = ((h - intrep) << 14)
		                / ((cropped_height - intrep) * 2);
	/* Vertical source positioning */
	dataformat = 2; /* Hard coded : we work with YUYV data */

	if (besviscal >= (1 << 14)) {
		/* If downscaling */
		origin_adress1 = (mga->bes_crop_top + ((offset_top * besviscal) >> 14))
			* bes_pitch * dataformat;
		
		origin_adress2 = (mga->bes_crop_top + ((offset_top * besviscal + subpixelcomp) >> 14))
			* bes_pitch * dataformat;

		weight1 = (offset_top * besviscal) & 0x3FFF;
		weight1s = 0;
		if (offset_top == 0) {
			weight2 = (- (1 << 13) - (besviscal & 0x3FFF)) & 0x3FFF;
			weight2s = 1;
		} else {
			weight2 = (offset_top * besviscal - (1 << 13)) & 0x3FFF;
			weight2s = 0;
		}
		tmp = mga->bes_crop_top + (offset_top * besviscal >> 14);
 		besv1srclst = win_height - 1 - tmp;
		besv1srcstp = (tmp & (1<<14)) ? 1 : 0;

		tmp = mga->bes_crop_top - ((offset_top * besviscal + subpixelcomp) >> 14);
		besv2srclst = win_height - 1 - tmp;
		besv2srcstp = (tmp & (1<<14)) ? 1 : 0;
	} else {
		/* If upscaling */
		origin_adress1 = (mga->bes_crop_top + ((offset_top * besviscal) >> 14))
			* bes_pitch * dataformat;
		/* If wrong things happen, check this first : */
		/* Should we use all digits of subpixelcomp ? */
		/* (here we truncate before multiplying by besviscal) */
		absval = abs((signed int)(offset_top - (subpixelcomp>>14))) * besviscal;
		origin_adress2 = (mga->bes_crop_top + (absval>>14)) * bes_pitch * dataformat;
		weight1 = (offset_top * besviscal) & 0x3FFF;
		weight1s = 0;
		/* Here : I don't know if weight2 is a 2's complement ... */
		/* 2's complement gives better results ... */
		if (offset_top * besviscal >= (1 << 13))
			weight2 = offset_top - (1 << 13);
		else
			weight2 = - (1 << 13) - offset_top * besviscal;
		weight2s = (weight2 & (1<<31)) ? 1 : 0;
		weight2 &= 0x3FFF;
		tmp = mga->bes_crop_top + (offset_top * besviscal >> 14);
		besv1srclst = win_height - 1 - tmp;
		besv1srcstp = (tmp & (1<<14)) ? 1 : 0;
		
		besv2srclst = win_height - 1 - (mga->bes_crop_top + (absval>>14));
		besv2srcstp = (((mga->bes_crop_top << 14) + absval) & (1 << 14)) ? 1 : 0;
	}
	/* Prepare BESCTL value */
	mga->besctl_value =
		1                                /* Enable preview */
		+ (besv1srcstp<<6)
		+ (besv2srcstp<<7)
		+ (BES_HFILTER_ON ? (1<<10) : 0)
		+ (BES_VFILTER_ON ? (1<<11) : 0)
		+ (BES_HFILTER_ON ? (1<<16) : 0) /* Chroma upsampling enabled */
		+ (BES_DITHER_ON ? (1<<18) : 0)  /* Dithering enabled */
		+ (BES_BW_ON ? (1<<20) : 0)      /* Preview is black and white */
		+ (0<<24)                        /* Software field select */
		+ (mga->besctl_value & (3<<25)); /* Window to display (keep the same) */
	/* Remember the origin_adress values */
	mga->bes_origin_topfield = origin_adress1;
	mga->bes_origin_bottomfield = origin_adress2;		
	/* We are ready to put the values in the mga registers */
	/* Disable preview */
	_mgawrite(mga, BESCTL, 0);
	/* Write values */
	_mgawrite(mga, BESGLOBCTL,
		(BES_ACC2XZOOM_ON ? 1 : 0) + (BES_HFILTER_ON ? 2 : 0));
	_mgawrite(mga, BESHCOORD, (x + w - offset_right - 1) + ((x + offset_left) << 16));
	_mgawrite(mga, BESHISCAL, beshiscal << 2);
	_mgawrite(mga, BESHSRCEND, beshsrcend << 2);
	_mgawrite(mga, BESHSRCLST, _VINWID - 1);
	_mgawrite(mga, BESHSRCST, beshsrcst << 2);
	_mgawrite(mga, BESPITCH, bes_pitch);
	_mgawrite(mga, BESV1SRCLST, besv1srclst);
	_mgawrite(mga, BESV2SRCLST, besv2srclst);
	_mgawrite(mga, BESV1WGHT, (weight1 << 2) + (weight1s << 16));
	_mgawrite(mga, BESV2WGHT, (weight2 << 2) + (weight2s << 16));
	_mgawrite(mga, BESVCOORD, (y + h - offset_bottom - 1) + ((y + offset_top) << 16));
	_mgawrite(mga, BESVISCAL, besviscal << 2);
	/* If preview was enabled, reenable it */
	if (mga->preview_active)
		_mgawrite(mga, BESCTL, mga->besctl_value);
}

void mgavideo_set_colorkey( struct mga_dev* mga, int r, int g, int b )
{
	switch(mga->desktop_depth) {
	case MGA_DEPTH15 : r >>= 3; g >>= 3; b >>= 3; break;
	case MGA_DEPTH16 : r >>= 3; g >>= 2; b >>= 3; break;
	/* Don't know what to do for other cases. */
 	}

#ifndef	NEVER
        writeb( XKEYOPMODE, mga->ctrl + PALWTADD );
        writeb( 0, mga->ctrl + X_DATAREG );
#else
        writeb( XKEYOPMODE, mga->ctrl + PALWTADD );
        writeb( 1, mga->ctrl + X_DATAREG );

        writeb( XCOLMSK0RED, mga->ctrl + PALWTADD );
        writeb( 0xff, mga->ctrl + X_DATAREG );

        writeb( XCOLMSK0GREEN, mga->ctrl + PALWTADD );
        writeb( 0xff, mga->ctrl + X_DATAREG );

        writeb( XCOLMSK0BLUE, mga->ctrl + PALWTADD );
        writeb( 0xff, mga->ctrl + X_DATAREG );

        writeb( XCOLKEY0RED, mga->ctrl + PALWTADD );
        writeb( r, mga->ctrl + X_DATAREG );

        writeb( XCOLKEY0GREEN, mga->ctrl + PALWTADD );
        writeb( g, mga->ctrl + X_DATAREG );

        writeb( XCOLKEY0BLUE, mga->ctrl + PALWTADD );
        writeb( b, mga->ctrl + X_DATAREG );
#endif
}

void mgavideo_set_overlay( struct mga_dev* mga )
{
        /* Activate the overlay mode */
        writeb( XKEYOPMODE, mga->ctrl + PALWTADD );
        writeb( 0, mga->ctrl + X_DATAREG );
}

void mgavideo_preview_enable( struct mga_dev* mga, int en )
{
        /* Activate scaler */
        if (en) {
                mga->preview_active = 1;
                _mgawrite(mga, BESCTL, mga->besctl_value);
	} else {
		mga->preview_active = 0;
                _mgawrite(mga, BESCTL, 0);
	}
}

/****************************************************************************
* initialization
****************************************************************************/
static void _mgavideo_init( struct mga_dev* mga )
{
        writeb( XMISCCTRL, mga->ctrl + PALWTADD );
        writeb( readb( mga->ctrl + X_DATAREG ) | 0x06, mga->ctrl + X_DATAREG );

        /* detect tv-tuner */
        _detect_tvt( mga );

        /* VIN programming changed : VIN windows are only programmed after vsync. */
        /* After reset, no data is written until a window is programmed. */
        /* Reset the VIN registers */
	_mgawrite( mga, VINCTL, 0);
	/* Enable grabbing (VINCTL0/1 = 0 -> no data grabbed for the moment) */
	if (mga->mga_type == MGA_G200)
      		_mgawrite( mga, VINCTL, 0x01010001);
	else
		_mgawrite( mga, VINCTL, 0x01010071); /* 1X mode selected */

        /* just in case, clear video interrupt */
        _mgawrite( mga, VICLEAR, VI_IVSYNC);

        /* Build VINCTL0/1 value (no VBI grabbing for the moment) */
	if (mga->mga_type == MGA_G400) {
        	mga->vinctl_value = ((_VINWID * 2) & (~7)) + 1;
	} else {
		mga->vinctl_value = (((_VINWID * 2) & (~7)) << 1) + 1;
	}

         /* Setup the video buffers */
        mga->buffer_stamp[0] = 0; mga->buffer_flags[0] = 0;
        mga->buffer_stamp[1] = 0; mga->buffer_flags[1] = 0;
        mga->buffer_stamp[2] = 0; mga->buffer_flags[2] = 0;
        mga->buffer_stamp[3] = 0; mga->buffer_flags[3] = 0;
        mga->field_req = FIELD_1 + FIELD_2;
        mga->active_buffer = 0;
        mga->next_active_buffer = 0;
        mga->vin_active = 0;

	if (mga->mga_type == MGA_G400) {  	/* G400 specific registers */
	/* set vbicount register */
		_mgawrite( mga, VBICOUNT, 0x00); /* already handled by ks0127 ? */
	/* setup vinheight register */
		_mgawrite( mga, VINHEIGHT, _VINHGT_PAL);
	}

	/* Additionnal data for the BES : desktop layout (hope X11 is running) */
	/* Read the desktop_depth */
	writeb(XMULCTRL, mga->ctrl + PALRDADD);
	mga->desktop_depth = readb(mga->ctrl + X_DATAREG);
	mga->desktop_width = 1024;
	mga->desktop_height = 768;
	mga->desktop_offsetx = 0;
	mga->desktop_offsety = 0;

        /* setup backend scaler */
	mga->preview_active = 0;

	/* Image cropping */
	mga->video_activelines = _VINHGT_PAL;
	mga->bes_crop_top = 0;
	mga->bes_crop_bottom = 0;
	mga->bes_crop_left = 0;
	mga->bes_crop_right = 0;

	/* Default preview mode */
	mga->interlaced_preview = 1;

        mgavideo_set_window(mga, 0, 0, _VINWID, _VINHGT);

	_mgawrite( mga, BESA1ORG, mga->vin_addr[0]+mga->bes_origin_topfield);
	_mgawrite( mga, BESA2ORG, mga->vin_addr[1]+mga->bes_origin_bottomfield);
	_mgawrite( mga, BESB1ORG, mga->vin_addr[0]+mga->bes_origin_topfield);
	_mgawrite( mga, BESB2ORG, mga->vin_addr[1]+mga->bes_origin_bottomfield);

        /* register interrupt routine */
        if( mga->irq ) {
                if( request_irq( mga->irq, interrupt_hw, SA_SHIRQ,
                                 mga->name, mga ) < 0 ) {
                        dprintk( "mgavideo: Denied IRQ %d\n", mga->irq );
                        mga->irq_status = 0;
                } else {
                        mga->irq_status = mga->irq;
                }
        }
}


#define NBOARDS 4
struct mga_dev mgavideo_dev[NBOARDS];

/* find all mga adapters */
struct mga_dev* mgavideo_get( void )
{
        int i;
        struct mga_dev* mga;

        for( i=0; i<NBOARDS; i++ )
        {
                mga = &mgavideo_dev[i];
                if( !mga->inuse && (mga->ctrl != NULL) )
                {
                        dprintk( "mgavideo: acquire video device %x\n", i );
			mga->inuse++;
                        MOD_INC_USE_COUNT;

                        _mgavideo_init( mga );
                        return mga;
                }
        }

        return NULL;
}

/* release a mga adapter that is not needed anymore */
void mgavideo_release( struct mga_dev* mga )
{
        if( mga == NULL )
                return;

        if( mga->ctrl )
        {
                /* clear any pending interrupt */
                _mgawrite( mga, VICLEAR, VI_IVSYNC );

                /* disable video interrupt */
		mga->interrupt_mask &= ~VI_IVSYNC;
                _mgawrite(mga, VIEN, _mgaread(mga, VIEN) & ~VI_IVSYNC );

                /* disable video-in */
		if (mga->mga_type)
                	_mgawrite( mga, VINCTL, 0x01010000 );
		else
			_mgawrite( mga, VINCTL, 0x01010030 );

                /* disable (temp) backend scaler */
                mgavideo_preview_enable( mga, 0 );

		/* Turn off the zr36060 - NOPE, this has to be done by the API driver */
		/* zr36060_deattach(&mga->zr); */

                /* un-register isr */
                if( mga->irq_status ) {
                        free_irq( mga->irq_status, mga );
                        mga->irq_status = 0;
                }

                MOD_DEC_USE_COUNT;
                mga->inuse--;
                dprintk( "mgavideo: release video device\n" );
                printk( "mgavideo: grabbed %d total frames\n", mga->grabbed );
        }
}

static void
mgavideo_init_hardware(struct pci_dev *dev, int found, int type)
{
	struct mga_dev	*mga;
	u32		mem;
	u32		addr = 0;

	if (dev->irq < 0) {
		printk("mgavideo: Must have an interrupt to work\n");
		printk("mgavideo: Turn on interrupt in bios\n");
		return;
	}

	if(found >= NBOARDS )
		return;

	/* the device is cool. set up a context for it */
	mga = &mgavideo_dev[found];
	memset( mga, 0, sizeof( *mga ) );

	mga->device = dev->device;
	mga->inuse = 0;

	mga->mga_type = type;

	/* This is setting up all the communication memory windows
	   _in_ the MGA card */
	/* read and map the frame buffer */
#if	LINUX_VERSION_CODE >= 0x020300
	addr = dev->resource[0].start;
#else
	addr = dev->base_address[0] & PCI_BASE_ADDRESS_MEM_MASK;
#endif
	mga->fb = ioremap( addr, memsize );

	/* Default pitch for mga */
	mga->vinpitch = _VINWID;
	
	/* setup address for video in memory */
	mem = memsize - 65000;

	/* Define 2 * 2 interlaced buffers */
	mem  = (mem - _VIN_WINDOW * 2) & VINADDRx_MASK;
	mga->vin_addr[2] = mem;
	mga->vin_addr[3] = mem + mga->vinpitch * 2;

	mem  = (mem - _VBI_WINDOW) & VBIADDRx_MASK;
	mem  = (mem - _VIN_WINDOW * 2) & VINADDRx_MASK;
	mga->vin_addr[0] = mem;
	mga->vin_addr[1] = mem + mga->vinpitch * 2;

	mem  = (mem - _VBI_WINDOW) & VBIADDRx_MASK;

	mga->vbi_addr[0] = mem;
	mem = (mem - _VBI_WINDOW) & VBIADDRx_MASK;
	mga->vbi_addr[1] = mem;
	mga->vbi_addr[2] = mem;
	mga->vbi_addr[3] = mem;


	/* JPEG_BUFSIZE uses to be 256k */
	/* allocate memory for the zr36060 JPEG frame buffer */
	dprintk("mgavideo: Allocating Zoran MJPEG buffers in MGA mem (size: %d kB)\n", 
		JPEG_BUFFSIZE / 1024);
	mem = (mem - JPEG_BUFFSIZE) & ~0x3ff;

	/* (1024 bytes for the command buffer) */
	mem = (mem - 1024) & ~0x3ff;
mem = 1024 * 1024 + 512 * 1024;
	/* the (relative) address seen for the CODEC (sits on the MGA) */
	mga->host_interface_base = mem;
	mga->interface_base= (u8 *)(mga->fb) + mem;
	mga->command_base = (u8 *)(mga->fb) + (mem + 256 * 1024);
	mga->read_base = mga->command_base + 512;
	mga->codec_offset = 0;
memset(mga->interface_base, 0, 256 * 1024);

	/* the virtual JPEG fb address seen from the CPU */
	mga->codecaddr = (u32)mga->codeccmd + 512;
	mga->codec_offset = 0;

	/* Default pitch for mga */
	mga->vinpitch = 704;

	/* read and map, the control aperature */
#if LINUX_VERSION_CODE >= 0x020300
	addr = dev->resource[1].start;
#else
	addr = dev->base_address[1] & PCI_BASE_ADDRESS_MEM_MASK;
#endif
	mga->ctrl = ioremap_nocache( addr, 0x4000 );

	/* fill in zr36060 register address */
	mga->zr36060.codec_id = ZR36060_ID;
	mga->zr36060.private = mga;
	mga->zr36060.memory_size = 8;

	mga->zr36060.input_count = 3;
	mga->zr36060.video_bus_size = 8;
	mga->zr36060.code_master = 0;
		
	mga->zr36060.codec_read_memory = mgacodec_read_memory;
	mga->zr36060.codec_write_memory = mgacodec_write_memory;
	mga->zr36060.codec_wakeup = mgacodec_zr36060_wakeup;
	mga->zr36060.codec_overlay = mgacodec_zr36060_overlay;
	mga->zr36060.codec_set_jpg = mgacodec_zr36060_set_jpg;
	mga->zr36060.codec_transfer = mgacodec_zr36060_transfer;
	mga->zr36060.codec_set_video = mgacodec_zr36060_set_video;

	mga->zr36060.norms = tvnorms;

	/* read the device's irq */
	mga->irq = dev->irq;

	/* reset the device */
	if (mga->mga_type == MGA_G200)
		_mgawrite( mga, VINCTL, 0x01010000 ); 
	else
		_mgawrite( mga, VINCTL, 0x01010030 );  /* G400 */

	/* setup the i2c */
	sprintf( mga->i2c.name, "mga%d", found );
	mga->i2c.id = I2C_BUSID_MGA;
	mga->i2c.data = mga;
#if LINUX_VERSION_CODE >= 0x020100
	mga->i2c.bus_lock = SPIN_LOCK_UNLOCKED;
#endif
	mga->i2c.attach_inform = _mgai2c_attach;
	mga->i2c.detach_inform = _mgai2c_detach;
	mga->i2c.i2c_setlines = _mgai2c_setlines;
	mga->i2c.i2c_getdataline = _mgai2c_getdataline;
	mga->i2c.i2c_read = NULL;

	mga->i2c.i2c_write = NULL;
	i2c_register_bus( &mga->i2c );

	sprintf( mga->name, "mgavideo%d", found );

	/* okay everything looks good */
	dprintk( "mgavideo: Found %s: device 0x%x irq %i\n", mga->name, mga->device, mga->irq );

}



/****************************************************************************
* linux kernel module api
****************************************************************************/
#ifdef MODULE
int init_module( void )
#else
int mgavideo_init( void )
#endif
{
        int found = 0;
        struct pci_dev* dev;

        dprintk( "mgavideo: Looking for boards...\n" );

        /* find the next MGA G100 or G200 */

        dev = NULL;
	while((dev = pci_find_device(PCI_VENDOR_ID_MATROX, PCI_DEVICE_ID_MATROX_G400, dev))) {
		mgavideo_init_hardware(dev, found++, MGA_G400);
	}


        dev = NULL;
	while((dev = pci_find_device(PCI_VENDOR_ID_MATROX, PCI_DEVICE_ID_MATROX_G200_AGP, dev))) {
		mgavideo_init_hardware(dev, found++, MGA_G200);
	}

        dev = NULL;
	while((dev = pci_find_device(PCI_VENDOR_ID_MATROX, PCI_DEVICE_ID_MATROX_G200_PCI, dev))) {
		mgavideo_init_hardware(dev, found++, MGA_G200);
	}

        dev = NULL;
	while((dev = pci_find_device(PCI_VENDOR_ID_MATROX, PCI_DEVICE_ID_MATROX_G100_MM, dev))) {
		mgavideo_init_hardware(dev, found++, MGA_G100);
	}


        dev = NULL;
	while((dev = pci_find_device(PCI_VENDOR_ID_MATROX, PCI_DEVICE_ID_MATROX_G100_AGP, dev))) {
		mgavideo_init_hardware(dev, found++, MGA_G100);
	}

        if( !found )
                return -ENODEV;

        return 0;
}

#ifdef MODULE
void cleanup_module( void )
{
        int i;
        struct mga_dev* mga;

        for( i=0; i<NBOARDS; i++ )
        {
                mga = &mgavideo_dev[i];
                if( mga->ctrl )
                {
                        i2c_unregister_bus( &mga->i2c );
                        iounmap( (void*)mga->ctrl );
                }
        }
}
#endif













